namespace recls
{
	using System;

	/// <summary>
	/// This class represents an entry in the file search
	/// </summary>
	/// <remarks>
	/// The class manages the recls library entry handle, and makes 
	/// accessible all the Win32 attributes of the structure via its properties
	/// </remarks>
	public class FileEntry
		: IDisposable
	{
		internal FileEntry(IntPtr entry)
		{
			m_entry = recls_api.CopyEntry(entry);
		}
		/// <summary>Releases the resources of the instance</summary>
		/// <remarks>The finaliser causes the release of the underlying <b>recls</b> library's resources</remarks>
 		~FileEntry()
		{
			Dispose();
		}

		/// <summary>
		/// Releases the recls library's resources used by this instance
		/// </summary>
		/// <value>This method does not return a value</value>
		/// <remarks>Calling <b>Dispose</b> allows for an early release of resources back to the <b>recls</b> library</remarks>
		public void Dispose()
		{
			recls_api.CloseEntry(ref m_entry);
		}

		/// <summary>
		/// The full path of the file entry
		/// </summary>
		/// <value>The full path of the file entry</value>
		/// <remarks>This represents the full path of the entry</remarks>
		public string Path
		{
			get
			{
				return recls_api.GetEntryPath(m_entry);
			}
		}

		/// <summary>
		/// The drive component of the file entry
		/// </summary>
		/// <value>The drive components of the file entry</value>
		/// <remarks>This represents the drive letter of the file entry</remarks>
		public char Drive
		{
			get
			{
				return recls_api.GetEntryDrive(m_entry);
			}
		}

		/// <summary>
		/// The directory component of the file entry
		/// </summary>
		/// <value>The directory component of the file entry</value>
		/// <remarks>This represents the directory of the entry</remarks>
		public string Directory
		{
			get
			{
				return recls_api.GetEntryDirectory(m_entry);
			}
		}

		/// <summary>
		/// The location of the file entry
		/// </summary>
		/// <value>The location of the file entry</value>
		/// <remarks>This represents the <see cref="Drive"/> + <see cref="Directory"/> of the given entry</remarks>
		public string DirectoryPath
		{
			get
			{
				return recls_api.GetEntryDirectoryPath(m_entry);
			}
		}

		/// <summary>
		/// The file component of the file entry
		/// </summary>
		/// <value>The file component of the file entry</value>
		/// <remarks>This represents the <see cref="FileName"/> + <see cref="FileExt"/> of the given entry</remarks>
		public string File
		{
			get
			{
				return recls_api.GetEntryFile(m_entry);
			}
		}

		/// <summary>
		/// The short version of the <see cref="File"/> property
		/// </summary>
		/// <value>The short (8.3) version of the <see cref="File"/> property</value>
		/// <remarks>This represents the short version of the <see cref="File"/> property. In cases where the <see cref="File"/>
		/// conforms to the 8.3 format, the value of this property is the empty string</remarks>
		public string ShortFile
		{
			get
			{
				return recls_api.GetEntryShortFile(m_entry);
			}
		}

		/// <summary>
		/// The name component of the file entry
		/// </summary>
		/// <value>The name component of the file entry</value>
		/// <remarks>The <see cref="File"/> property, without the extension (<see cref="FileExt"/>)</remarks>
		public string FileName
		{
			get
			{
				return recls_api.GetEntryFileName(m_entry);
			}
		}

		/// <summary>
		/// The extension component of the file entry
		/// </summary>
		/// <value>The extension component of the file entry</value>
		/// <remarks>
		/// <p>
		/// The <see cref="File"/> property, without the name (<see cref="FileName"/>)
		/// </p>
		/// <p>
		/// <b>Note:</b> the value of this property does <b>not</b> contain the <c>.</c> separator.
		/// </p>
		/// <p>
		/// <b>Note:</b> if the entry does not have an extension, the value of this property is the empty string
		/// </p>
		/// </remarks>
		public string FileExt
		{
			get
			{
				return recls_api.GetEntryFileExt(m_entry);
			}
		}

		/// <summary>
		/// The creation time of the entry
		/// </summary>
		/// <value>The creation time of the entry</value>
		/// <remarks>This property represents the time that the entry was created</remarks>
		/// <seeAlso>DateTime</seeAlso>
		public DateTime CreationTime
		{
			get
			{
				return recls_api.GetEntryCreationTime(m_entry);
			}
		}

		/// <summary>
		/// The modification time of the entry
		/// </summary>
		/// <value>The modification time of the entry</value>
		/// <remarks>This property represents the time that the entry was created</remarks>
		/// <seeAlso>DateTime</seeAlso>
		public DateTime ModificationTime
		{
			get
			{
				return recls_api.GetEntryModificationTime(m_entry);
			}
		}

		/// <summary>
		/// The last access time of the entry
		/// </summary>
		/// <value>The last access time of the entry</value>
		/// <remarks>This property represents the time that the entry was last accessed</remarks>
		/// <seeAlso>DateTime</seeAlso>
		public DateTime LastAccessTime
		{
			get
			{
				return recls_api.GetEntryLastAccessTime(m_entry);
			}
		}

		/// <summary>
		/// The last status change time of the entry
		/// </summary>
		/// <value>The last status change time of the entry</value>
		/// <remarks>This property represents the time that the entry was last changed</remarks>
		/// <seeAlso>DateTime</seeAlso>
		public DateTime LastStatusChangeTime
		{
			get
			{
				return recls_api.GetEntryLastStatusChangeTime(m_entry);
			}
		}

		/// <summary>
		/// The size of the file entry
		/// </summary>
		/// <value>The size of the file entry</value>
		/// <remarks>This property represents the size of the entry, if it is a file. Other types return an entry size of <c>0</c></remarks>
		public ulong Size
		{
			get
			{
				return recls_api.GetEntrySize(m_entry);
			}
		}

		/// <summary>
		/// Indicates whether the entry is readonly
		/// </summary>
		/// <value>Indicates whether the entry is readonly</value>
		/// <remarks>This property indicates whether the entry is readonly. Files and directories express this property, but other types may not</remarks>
		public bool IsReadOnly
		{
			get
			{
				return recls_api.GetEntryIsReadOnly(m_entry);
			}
		}

		/// <summary>
		/// Indicates whether the item is a directory
		/// </summary>
		/// <value>Indicates whether the entry is a directory</value>
		/// <remarks>This property indicates whether the entry is a directory. This returns <b>true</b> for true directories and for directory links.</remarks>
		public bool IsDirectory
		{
			get
			{
				return recls_api.GetEntryIsDirectory(m_entry);
			}
		}

		/// <summary>
		/// Indicates whether the item is a link
		/// </summary>
		/// <value>Indicates whether the entry is a link</value>
		/// <remarks>
		/// <p>
		/// This property indicates whether the entry is a link.
		/// </p>
		/// <p>
		/// <b>Note:</b> links are not currently detected by the recls library for the Win32 platform
		/// </p>
		/// </remarks>
		public bool IsLink
		{
			get
			{
				return recls_api.GetEntryIsLink(m_entry);
			}
		}

		/// <summary>
		/// The directory parts for the entry
		/// </summary>
		/// <value>The individual parts of the entry's directory</value>
		/// <remarks>This member contains an enumerable sequence of the parts of the directory if the search was carried out using the <see cref="RECLS_FLAG.RECLS_F_DIRECTORY_PARTS"/> member of the <see cref="RECLS_FLAG"/> enumeration</remarks>
		public DirectoryParts DirectoryParts
		{
			get
			{
				if(null == m_dirParts)
				{
					m_dirParts = new DirectoryParts(m_entry);
				}

				return m_dirParts;
			}
		}

		/// <summary>
		///  The entry
		/// </summary>
		private IntPtr			m_entry;
		/// <summary>
		/// The directory parts object. Cached since it never changes.
		/// </summary>
		private DirectoryParts	m_dirParts;

		/// <summary>
		/// Provides a string representation of the file entry
		/// </summary>
		/// <returns>The full path of the entry</returns>
		/// <remarks>This method provides an implicit conversion to string form, such that an entry
		/// reference may be used without recourse to the Path property</remarks>
		public override string ToString()
		{
			return Path;
		}
	}
}
