namespace recls
{
	using System;
	using System.Runtime.InteropServices;
	using System.Text;
	using hrecls_t = System.IntPtr;
	using recls_info_t = System.IntPtr;
	using recls_filesize_t = System.UInt64;

	/// <summary>
	/// recls search flags
	/// </summary>
	/// <remarks>
	/// <p>
	/// These flags alter the behaviour of a given search. Some of the flags affect each other, such
	/// as <see cref="RECLS_F_FILES"/>, <see cref="RECLS_F_DIRECTORIES"/>, <see cref="RECLS_F_LINKS"/> and
	/// <see cref="RECLS_F_DEVICES"/>, which together encompass the possible types of file entries retrieved
	/// by a search.
	/// </p>
	/// <p>
	/// <b>Note:</b> if none of these flags are specified, <see cref="RECLS_F_FILES"/> will be assumed
	/// </p>
	/// <p>
	/// <b>Note:</b> <see cref="RECLS_F_LINKS"/> and <see cref="RECLS_F_DEVICES"/> are not currently supported
	/// </p>
	/// <p>
	/// The other flags moderate how the search is conducted:
	/// </p>
	/// <p>
	/// <see cref="RECLS_F_RECURSIVE"/> causes the search to be conducted on the given directory, and all its sub-directories
	/// </p>
	/// <p>
	/// <see cref="RECLS_F_NO_FOLLOW_LINKS"/> is not currently supported
	/// </p>
	/// <p>
	/// <see cref="RECLS_F_DIRECTORY_PARTS"/> causes the components of the directory for each entry to be calculated, which are
	/// then available as the <see cref="FileEntry.DirectoryParts"/> property
	/// </p>
	/// <p>
	/// <see cref="RECLS_F_DETAILS_LATER"/> is not currently supported
	/// </p>
	/// </remarks>
	[Flags]
	public enum RECLS_FLAG
	{
			/// <summary>
			/// Include files in search. Included by default if none specified
			/// </summary>
			RECLS_F_FILES               =   0x00000001
		,	/// <summary>
			/// Include directories in search
			/// </summary>
			RECLS_F_DIRECTORIES         =   0x00000002
		,	/// <summary>
			/// Include links in search
			/// </summary>
		    RECLS_F_LINKS				=   0x00000004
		,	/// <summary>
			/// Include devices in search
			/// </summary>
			RECLS_F_DEVICES             =   0x00000008
		,	/// <summary>
			/// Searches given directory and all sub-directories
			/// </summary>
			RECLS_F_RECURSIVE           =   0x00010000
		,	/// <summary>
			/// Does not expand links
			/// </summary>
			RECLS_F_NO_FOLLOW_LINKS     =   0x00020000
		,	/// <summary>
			/// Fills out the directory parts
			/// </summary>
			RECLS_F_DIRECTORY_PARTS     =   0x00040000
		,	/// <summary>
			/// Does not fill out anything other than the path
			/// </summary>
		   RECLS_F_DETAILS_LATER		=   0x00080000
	};

	/// <summary>
	/// The exception type used by the recls API
	/// </summary>
	/// <remarks>Instances of this exception are throw when error conditions are encountered. They are 
	/// not thrown when a search returns no items, since that is a valid search result.</remarks>
	public class ReclsException
		: ApplicationException
	{
		/// <summary>
		/// Creates an instance of the exception
		/// </summary>
		internal ReclsException(string message, int code)
			: base(_make_x_str(message, code))
		{}

		private static string _make_x_str(string message, int code)
		{
			if(code == recls_api.RECLS_RC_OK)
			{
				return message;
			}
			else
			{
				return message + ": " + recls_api.GetErrorString(code) + " (" + code + ")";
			}
		}
	}

	/// <summary>
	/// This class wraps access to the raw recls API
	/// </summary>
	internal sealed class recls_api
	{
		private recls_api()
		{}

		/// <summary>
		/// General success code
		/// </summary>
		public const int	RECLS_RC_OK						=	(0);
		/// <summary>
		/// General failure code
		/// </summary>
		public const int	RECLS_RC_FAIL                   =	(-1);

		/// <summary>
		/// No search is currently active
		/// </summary>
		public const int	RECLS_RC_SEARCH_NO_CURRENT      =	((-1 - 1001));
		/// <summary>
		/// The directory was invalid, or does not exist
		/// </summary>
		public const int	RECLS_RC_INVALID_DIRECTORY      =	((-1 - 1002));
		/// <summary>
		/// No more data is available
		/// </summary>
		public const int	RECLS_RC_NO_MORE_DATA           =	((-1 - 1003));
		/// <summary>
		/// Memory exhaustion
		/// </summary>
		public const int	RECLS_RC_OUT_OF_MEMORY          =	((-1 - 1004));
		/// <summary>
		/// Function not implemented
		/// </summary>
		public const int	RECLS_RC_NOT_IMPLEMENTED        =	((-1 - 1005));

		public static hrecls_t OpenSearch(string searchRoot, string pattern, RECLS_FLAG flags)
		{
			hrecls_t	hSrch;
			int 		code = Recls_Search(searchRoot, pattern, (uint)flags, out hSrch);

			if(code == RECLS_RC_NO_MORE_DATA)
			{
				hSrch = hrecls_t.Zero;
			}
			else if(code < 0)
			{
				throw new ReclsException("Failed to create search", code);
			}

			return hSrch;
		}

		public static hrecls_t OpenFtpSearch(string host, string username, string password, string searchRoot, string pattern, RECLS_FLAG flags)
		{
			hrecls_t	hSrch;
			int 		code = Recls_SearchFtp(host, username, password, searchRoot, pattern, (uint)flags, out hSrch);

			if(code == RECLS_RC_NO_MORE_DATA)
			{
				hSrch = hrecls_t.Zero;
			}
			else if(code < 0)
			{
				throw new ReclsException("Failed to create search", code);
			}

			return hSrch;
		}

		public static void CloseSearch(ref hrecls_t hSrch)
		{
			if(IntPtr.Zero != hSrch)
			{
				Recls_SearchClose(hSrch);
				hSrch = IntPtr.Zero;
			}
		}

		public static string GetErrorString(int code)
		{
			StringBuilder	buffer		=	new StringBuilder(261);
			uint			capacity	=	(uint)buffer.Capacity;
			int				cch 		= 	Recls_GetErrorString(code, buffer, capacity);

			buffer.Length = cch;

			return buffer.ToString();
		}

		public static void CloseEntry(ref recls_info_t entry)
		{
			if(IntPtr.Zero != entry)
			{
				Recls_CloseDetails(entry);
				entry = IntPtr.Zero;
			}
		}

		public static recls_info_t CopyEntry(recls_info_t entry)
		{
			recls_info_t	infoCopy;
			int				code	=	Recls_CopyDetails(entry, out infoCopy);

			if(code < 0)
			{
				throw new ReclsException("Failed to copy details", code);
			}

			return infoCopy;
		}

		public static bool NextEntry(hrecls_t hSrch, out recls_info_t entry)
		{
			int code = Recls_GetNextDetails(hSrch, out entry);

			if(code == RECLS_RC_NO_MORE_DATA)
			{
				return false;
			}
			else if(code < 0)
			{
				throw new ReclsException("Failed to get next entry", code);
			}
			else
			{
				return true;
			}
		}

		public static void GetEntry(hrecls_t hSrch, out recls_info_t entry)
		{
			int code = Recls_GetDetails(hSrch, out entry);

			if(code < 0)
			{
				throw new ReclsException("Failed to create search", code);
			}
		}

		/// <summary>
		/// A delegate defining the various string property access functions
		/// </summary>
		private delegate uint Recls_GetStringProperty_dg(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		private static String GetEntryString(recls_info_t entry, Recls_GetStringProperty_dg dg)
		{
			uint			cch 		= 	dg(entry, null, 0);
			StringBuilder	buffer		=	new StringBuilder(1 + (int)cch);
			uint			capacity	=	(uint)buffer.Capacity;

			cch = dg(entry, buffer, capacity);

			buffer.Length = (int)cch;

			return buffer.ToString();
		}

		public static String GetEntryPath(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetPathProperty));
		}

		public static char GetEntryDrive(recls_info_t entry)
		{
			char	chDrive;

			Recls_GetDriveProperty(entry, out chDrive);

			return chDrive;
		}

		public static String GetEntryDirectory(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetDirectoryProperty));
		}

		public static String GetEntryDirectoryPath(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetDirectoryPathProperty));
		}

		public static String GetEntryFile(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetFileProperty));
		}

		public static String GetEntryShortFile(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetShortFileProperty));
		}

		public static String GetEntryFileName(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetFileNameProperty));
		}

		public static String GetEntryFileExt(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetFileExtProperty));
		}

		public static int GetEntryDirectoryPartsCount(recls_info_t entry)
		{
			return (int)Recls_GetDirectoryPartProperty(entry, -1, null, 0);
		}

		public static string GetEntryDirectoryPart(recls_info_t entry, int index)
		{
			StringBuilder	buffer		=	new StringBuilder(261);
			uint			capacity	=	(uint)buffer.Capacity;
			uint			cch 		= 	Recls_GetDirectoryPartProperty(entry, index, buffer, capacity);

			buffer.Length = (int)cch;

			return buffer.ToString();
		}

		public static DateTime GetEntryCreationTime(recls_info_t entry)
		{
			return DateTime.FromFileTime(Recls_GetCreationTime(entry));
		}
		public static DateTime GetEntryModificationTime(recls_info_t entry)
		{
			return DateTime.FromFileTime(Recls_GetModificationTime(entry));
		}
		public static DateTime GetEntryLastAccessTime(recls_info_t entry)
		{
			return DateTime.FromFileTime(Recls_GetLastAccessTime(entry));
		}
		public static DateTime GetEntryLastStatusChangeTime(recls_info_t entry)
		{
			return DateTime.FromFileTime(Recls_GetLastStatusChangeTime(entry));
		}

		public static ulong GetEntrySize(recls_info_t entry)
		{
			recls_filesize_t    size;

			Recls_GetSizeProperty(entry, out size);

			return size;
		}

		public static bool GetEntryIsReadOnly(recls_info_t entry)
		{
			return 0 != Recls_IsFileReadOnly(entry);
		}

		public static bool GetEntryIsDirectory(recls_info_t entry)
		{
			return 0 != Recls_IsFileDirectory(entry);
		}

		public static bool GetEntryIsLink(recls_info_t entry)
		{
			return 0 != Recls_IsFileLink(entry);
		}

		/// <summary>
		/// Starts a search in the given root, for the given pattern according to the given flags
		/// </summary>
		[DllImport("recls_dll", EntryPoint = "Recls_Search", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_Search(string searchRoot, string pattern, uint flags, out hrecls_t hSrch);

		[DllImport("recls_dll", EntryPoint = "Recls_SearchFtp", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_SearchFtp(string host, string username, string password, string searchRoot, string pattern, uint flags, out hrecls_t hSrch);

		[DllImport("recls_dll", EntryPoint = "Recls_SearchClose", CallingConvention=CallingConvention.Cdecl, ExactSpelling=true)]
		private static extern void Recls_SearchClose(hrecls_t hSrch);

		[DllImport("recls_dll", EntryPoint = "Recls_GetErrorString", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_GetErrorString(int code, StringBuilder buffer, uint cchBuffer);

		[DllImport("recls_dll", EntryPoint = "Recls_GetNextDetails", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_GetNextDetails(hrecls_t hSrch, out recls_info_t info);

		[DllImport("recls_dll", EntryPoint = "Recls_GetDetails", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_GetDetails(hrecls_t hSrch, out recls_info_t info);

		[DllImport("recls_dll", EntryPoint = "Recls_CopyDetails", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_CopyDetails(recls_info_t fileInfo, out recls_info_t infoCopy);

		[DllImport("recls_dll", EntryPoint = "Recls_CloseDetails", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_CloseDetails(recls_info_t fileInfo);

		[DllImport("recls_dll", EntryPoint = "Recls_GetPathProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetPathProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport("recls_dll", EntryPoint = "Recls_GetDriveProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern void Recls_GetDriveProperty(recls_info_t fileInfo, out char chDrive);

		[DllImport("recls_dll", EntryPoint = "Recls_GetDirectoryProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetDirectoryProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport("recls_dll", EntryPoint = "Recls_GetDirectoryPathProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetDirectoryPathProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport("recls_dll", EntryPoint = "Recls_GetShortFileProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetShortFileProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport("recls_dll", EntryPoint = "Recls_GetFileProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetFileProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport("recls_dll", EntryPoint = "Recls_GetFileNameProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetFileNameProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport("recls_dll", EntryPoint = "Recls_GetFileExtProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetFileExtProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport("recls_dll", EntryPoint = "Recls_GetCreationTime", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern long Recls_GetCreationTime(recls_info_t fileInfo);
		[DllImport("recls_dll", EntryPoint = "Recls_GetModificationTime", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern long Recls_GetModificationTime(recls_info_t fileInfo);
		[DllImport("recls_dll", EntryPoint = "Recls_GetLastAccessTime", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern long Recls_GetLastAccessTime(recls_info_t fileInfo);
		[DllImport("recls_dll", EntryPoint = "Recls_GetLastStatusChangeTime", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern long Recls_GetLastStatusChangeTime(recls_info_t fileInfo);

		[DllImport("recls_dll", EntryPoint = "Recls_GetSizeProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern void Recls_GetSizeProperty(recls_info_t fileInfo, out recls_filesize_t size);

		[DllImport("recls_dll", EntryPoint = "Recls_GetDirectoryPartProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetDirectoryPartProperty(recls_info_t fileInfo, int part, StringBuilder buffer, uint cchBuffer);

		[DllImport("recls_dll", EntryPoint = "Recls_IsFileReadOnly", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_IsFileReadOnly(recls_info_t fileInfo);

		[DllImport("recls_dll", EntryPoint = "Recls_IsFileDirectory", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_IsFileDirectory(recls_info_t fileInfo);

		[DllImport("recls_dll", EntryPoint = "Recls_IsFileLink", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_IsFileLink(recls_info_t fileInfo);
	}

	// This class serves as the namespace summary doc for NDoc, because C# cannot use <summary> tags on namespaces
	/// <summary>
	/// Contains the recls C# mapping classes and enumerations
	/// </summary>
	class NamespaceDoc
	{}
}
