/* /////////////////////////////////////////////////////////////////////////////
 * File:        ReclsFtpSearchDirectoryNode.cpp
 *
 * Purpose:     Implementation of the ReclsFtpSearchDirectoryNode class for Win32.
 *
 * Created:     1st June 2004
 * Updated:     6th March 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright 2004-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include "recls.h"

// There's a really weird bug with Intel on Win32 (with MSVC libs), such that
// typeinfo.h gets included and attempts to derive exception classes in the global
// namespace from the (undefined) ::exception
#if defined(RECLS_COMPILER_IS_INTEL) && \
	defined(RECLS_PLATFORM_IS_WIN32)
//# define _INC_TYPEINFO
//# include <exception>
#endif /* Intel on Win32 */

#include "recls_internal.h"
#include "recls_assert.h"
#include "recls_util.h"

#if !defined(RECLS_COMPILER_IS_MSVC) || \
    _MSC_VER >= 1200
# define USE_SEARCHSPEC_SEQ
#endif /* compiler */

#include "recls_debug.h"

#if defined(RECLS_DELAY_LOAD_WININET)
# ifdef INETSTL_INCL_H_INETSTL
#  error INETSTL_INCL_H_INETSTL
# endif /* INETSTL_INCL_H_INETSTL */
# include "recls_wininet_dl.h"
#endif /* RECLS_DELAY_LOAD_WININET */

#include "ReclsFtpSearchDirectoryNode_win32.h"
#include "EntryFunctions.h"

#if defined(RECLS_CHAR_TYPE_IS_WCHAR)
# include <winstl_char_conversions.h>
#endif /* RECLS_CHAR_TYPE_IS_???? */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

namespace {

#if defined(RECLS_COMPILER_IS_COMO) || \
    defined(RECLS_COMPILER_IS_DMC) || \
    defined(RECLS_COMPILER_IS_GCC) || \
    defined(RECLS_COMPILER_IS_INTEL) || \
    defined(RECLS_COMPILER_IS_MWERKS)
typedef ReclsFtpSearchDirectoryNode::traits_type    traits_type;
#else /* ? compiler */
using ReclsFtpSearchDirectoryNode::traits_type;
#endif /* compiler */

stlsoft_ns_using(c_str_ptr)

} // anonymous namespace

/* /////////////////////////////////////////////////////////////////////////////
 * Constants
 */

//static const recls_char_t s_defaultSearchRoot[]       =   RECLS_LITERAL(".");
static const recls_char_t   s_localDir[]                =   RECLS_LITERAL(".");
//static const recls_char_t s_parentDir[]               =   RECLS_LITERAL("..");

//static const recls_char_t s_protocolSep[]             =   RECLS_LITERAL("://");
//static const recls_char_t s_protocolFile[]            =   RECLS_LITERAL("file");
//static const recls_char_t s_protocolFTP[]             =   RECLS_LITERAL("ftp");
//static const recls_char_t s_protocolWin32Registry[]   =   RECLS_LITERAL("win/registry");

/* /////////////////////////////////////////////////////////////////////////////
 * Utility functions
 */

inline size_t count_dir_parts(recls_char_t const *begin, recls_char_t const *end)
{
    return count_char_instances(begin, end, traits_type::path_name_separator());
}

/// Creates an entry from the platform-specific record

static recls_info_t create_entryinfo_from_psrecord(recls_char_t const *rootDir, recls_uint32_t flags, traits_type::find_data_type const &findData)
{
    typedef recls_char_t    char_type;

    function_scope_trace("create_entryinfo_from_psrecord");

    // size of structure is:
    //
    //    offsetof(struct recls_fileinfo_t, data)
    //  + directory parts
    //  + full path (+ null)
    //  + short name (+ null)

    // Bizarrely, the findData.cFileName can come in either as the file name or as the full path,
    // which means that we're in for some effort to handle that.

    int     isRooted    =   findData.cFileName[0] == '/';
    size_t  cchRootDir  =   traits_type::str_len(rootDir);
    size_t  cchEntry    =   traits_type::str_len(findData.cFileName);
    size_t  cchFile     =   isRooted ? (cchEntry - ((traits_type::str_rchr(findData.cFileName, '/') + 1) - &findData.cFileName[0])) : cchEntry;
    size_t  cchFullPath =   isRooted ? cchEntry : (cchRootDir + cchFile);

    recls_assert(cchRootDir > 0);
    recls_assert(rootDir[cchRootDir - 1] == traits_type::path_name_separator());

    size_t  cDirParts   =   ((flags & RECLS_F_DIRECTORY_PARTS) == RECLS_F_DIRECTORY_PARTS) ? count_dir_parts(rootDir, rootDir + cchRootDir) : 0;
    size_t  cbPath      =   align_up_size(sizeof(char_type) * (1 + cchFullPath));
    size_t  cbAlt       =   align_up_size(sizeof(char_type) * (1 + RECLS_NUM_ELEMENTS(findData.cAlternateFileName)));
    size_t  cb          =   offsetof(struct recls_fileinfo_t, data)
                        +   cDirParts * sizeof(recls_strptrs_t)
                        +   cbPath
                        +   cbAlt;

    struct recls_fileinfo_t *info   =   const_cast<struct recls_fileinfo_t*>(Entry_Allocate(cb));

    if(NULL != info)
    {
        char_type               *fullPath   =   (char_type*)&info->data[cDirParts * sizeof(recls_strptrs_t)];
        char_type               *altName    =   (char_type*)&info->data[cDirParts * sizeof(recls_strptrs_t) + cbPath];

        // full path
        traits_type::str_copy(fullPath, isRooted ? RECLS_LITERAL("") : rootDir);
        traits_type::str_cat(fullPath, findData.cFileName);
        info->path.begin                    =   fullPath;
        info->path.end                      =   fullPath + cchFullPath;

        debug_printf("entry: %s\n", fullPath);

        // drive, directory, file (name + ext)
        info->drive                         =   '\0';
        info->directory.begin               =   fullPath;
        info->directory.end                 =   fullPath + cchRootDir;
        info->fileName.begin                =   info->directory.end;
#if defined(RECLS_CHAR_TYPE_IS_CHAR)
        info->fileName.end                  =   strrchr(info->directory.end, '.');
#elif defined(RECLS_CHAR_TYPE_IS_WCHAR)
        info->fileName.end                  =   wcsrchr(info->directory.end, L'.');
#else
# error 
#endif /* RECLS_CHAR_TYPE_IS_???? */

        if(NULL != info->fileName.end)
        {
            info->fileExt.begin             =   info->fileName.end + 1;
            info->fileExt.end               =   info->directory.end +  cchFile;
        }
        else
        {
            info->fileName.end              =   info->directory.end +  cchFile;
            info->fileExt.begin             =   info->directory.end +  cchFile;
            info->fileExt.end               =   info->directory.end +  cchFile;
        }

        // determine the directory parts
        char_type const         *p          =   info->directory.begin;
        char_type const         *l          =   info->directory.end;
        struct recls_strptrs_t  *begin      =   (struct recls_strptrs_t*)&info->data[0];

        info->directoryParts.begin          =   begin;
        info->directoryParts.end            =   begin + cDirParts;

        if(info->directoryParts.begin != info->directoryParts.end)
        {
            recls_assert((flags & RECLS_F_DIRECTORY_PARTS) == RECLS_F_DIRECTORY_PARTS);

            begin->begin = p;

            for(; p != l; ++p)
            {
                if(*p == traits_type::path_name_separator())
                {
                    begin->end = p + 1;

                    if(++begin != info->directoryParts.end)
                    {
                        begin->begin = p + 1;
                    }
                }
            }
        }

        // alt name
        traits_type::str_copy(altName, findData.cAlternateFileName);
        info->shortFile.begin   =   altName;
        info->shortFile.end     =   altName + traits_type::str_len(altName);

        // attributes
        info->attributes        =   findData.dwFileAttributes;

        // time, size
        info->creationTime      =   findData.ftCreationTime;
        info->modificationTime  =   findData.ftLastWriteTime;
        info->lastAccessTime    =   findData.ftLastAccessTime;
        info->size.u.HighPart   =   findData.nFileSizeHigh;
        info->size.u.LowPart    =   findData.nFileSizeLow;

        // Checks
        recls_assert(info->path.begin < info->path.end);

        recls_assert(info->directory.begin < info->directory.end);
        recls_assert(info->path.begin <= info->directory.begin);
        recls_assert(info->directory.end <= info->path.end);

        recls_assert(info->fileName.begin <= info->fileName.end);

        recls_assert(info->fileExt.begin <= info->fileExt.end);
        recls_assert(traits_type::str_len(info->fileExt.begin) == static_cast<size_t>(info->fileExt.end - info->fileExt.begin));

        recls_assert(info->fileName.begin < info->fileExt.end);
        recls_assert(info->fileName.end <= info->fileExt.begin);

        recls_assert(info->directory.end == info->fileName.begin);

//      recls_assert(traits_type::str_len(info->path.begin) == (info->
    }

    return info;
}

/* /////////////////////////////////////////////////////////////////////////////
 * ReclsFtpSearchDirectoryNode
 */

ReclsFtpSearchDirectoryNode::ReclsFtpSearchDirectoryNode(HINTERNET connection, recls_uint32_t flags, ReclsFtpSearchDirectoryNode::char_type const *rootDir, ReclsFtpSearchDirectoryNode::char_type const *pattern)
    : m_connection(connection)
    , m_current(NULL)
    , m_dnode(NULL)
    , m_flags(flags)
{
    function_scope_trace("ReclsFtpSearchDirectoryNode::ReclsFtpSearchDirectoryNode");

    recls_assert(NULL != m_connection);

    // (1) Do the directories first
    if(flags & RECLS_F_RECURSIVE)
    {
        file_find_sequence_type directories(connection, rootDir, traits_type::pattern_all(), file_find_sequence_type::directories);

        file_find_sequence_type::const_iterator begin   =   directories.begin();
        file_find_sequence_type::const_iterator end     =   directories.end();

        for(; begin != end; ++begin)
        {
            m_directories.push_back((*begin).get_path());
        }
    }
    m_directoriesBegin = m_directories.begin(); // This is == end() in non-recursive cases

    // (2) Then the entries
#ifdef USE_SEARCHSPEC_SEQ
    m_entries       =   entry_sequence_type(connection, rootDir, pattern, traits_type::path_separator(), _ssFlags_from_reclsFlags(flags));
#else /* ? USE_SEARCHSPEC_SEQ */
    m_entries       =   entry_sequence_type(connection, rootDir, pattern, _ssFlags_from_reclsFlags(flags));
#endif /* USE_SEARCHSPEC_SEQ */
    m_entriesBegin  =   m_entries.begin();

    if(0 == traits_type::str_compare(s_localDir, pattern))
    {
        // This special case handles "." as a search pattern.
        string_type temp(rootDir);

        if( !temp.empty() &&
            temp.back() == '/')
        {
            temp = string_type(&temp[0], &temp[temp.length() - 1]);
        }

        size_t              cch =   traits_type::get_full_path_name(m_connection, c_str_ptr(temp), 0, NULL);
        string_type         temp2(cch, ' ');
        recls_char_t        *file;

        traits_type::get_full_path_name(m_connection, c_str_ptr(temp), 1 + temp2.length(), &temp2[0], &file);

        m_rootDir.assign(&temp2[0], file);
        m_pattern.assign(file);
    }
    else
    {
        m_rootDir = rootDir;
        if( !m_rootDir.empty() &&
            m_rootDir.back() != '/')
        {
            m_rootDir += '/';
        }
        m_pattern = pattern;
    }

    recls_assert(stlsoft_raw_offsetof(ReclsFtpSearchDirectoryNode, m_entries) < stlsoft_raw_offsetof(ReclsFtpSearchDirectoryNode, m_entriesBegin));
    recls_assert(stlsoft_raw_offsetof(ReclsFtpSearchDirectoryNode, m_directories) < stlsoft_raw_offsetof(ReclsFtpSearchDirectoryNode, m_directoriesBegin));
}

/* static */ ReclsFtpSearchDirectoryNode *ReclsFtpSearchDirectoryNode::FindAndCreate(   HINTERNET                                       connection
                                                                                    ,   recls_uint32_t                                  flags
                                                                                    ,   ReclsFtpSearchDirectoryNode::char_type const    *rootDir
                                                                                    ,   ReclsFtpSearchDirectoryNode::char_type const    *pattern)
{
    ReclsFtpSearchDirectoryNode  *node;

    function_scope_trace("ReclsFtpSearchDirectoryNode::FindAndCreate");

#ifdef RECLS_COMPILER_THROWS_ON_NEW_FAIL
    try
    {
#endif /* RECLS_COMPILER_THROWS_ON_NEW_FAIL */
        node = new ReclsFtpSearchDirectoryNode(connection, flags, rootDir, pattern);
#ifdef RECLS_COMPILER_THROWS_ON_NEW_FAIL
    }
    catch(std::bad_alloc &)
    {
        node = NULL;
    }
#endif /* RECLS_COMPILER_THROWS_ON_NEW_FAIL */

    if(NULL != node)
    {
        // Ensure that it, or one of its sub-nodes, has matching entries.
        recls_rc_t  rc = node->Initialise();

        if(RECLS_FAILED(rc))
        {
            delete node;

            node = NULL;
        }
    }

    recls_assert(NULL == node || node->is_valid_());

    return node;
}

/* static */ ReclsFtpSearchDirectoryNode *ReclsFtpSearchDirectoryNode::FindAndCreate(   HINTERNET                                       connection
                                                                                    ,   recls_uint32_t                                  flags
                                                                                    ,   ReclsFtpSearchDirectoryNode::char_type const    *rootDir
                                                                                    ,   ReclsFtpSearchDirectoryNode::char_type const    *subDir
                                                                                    ,   ReclsFtpSearchDirectoryNode::char_type const    *pattern)
{
    recls_assert(rootDir[traits_type::str_len(rootDir) - 1] == traits_type::path_name_separator());

    string_type     compositeDir(rootDir);

    compositeDir += subDir;

    return FindAndCreate(connection, flags, c_str_ptr(compositeDir), pattern);
}

ReclsFtpSearchDirectoryNode::~ReclsFtpSearchDirectoryNode()
{
    function_scope_trace("ReclsFtpSearchDirectoryNode::~ReclsFtpSearchDirectoryNode");

    Entry_Release(m_current);

    delete m_dnode;
}

recls_rc_t ReclsFtpSearchDirectoryNode::Initialise()
{
    function_scope_trace("ReclsFtpSearchDirectoryNode::Initialise");

    recls_rc_t  rc;

    recls_assert(NULL == m_current);
    recls_assert(NULL == m_dnode);

    if(m_entriesBegin != m_entries.end())
    {
        // (i) Try getting a file first,
#if 0
        // The way glob_sequence works
//      m_current = create_entryinfo_from_psrecord(m_rootDir, m_flags, *m_entriesBegin);
#else /* ? 0 */
        // The way basic_findfile_sequence<> works
        m_current = create_entryinfo_from_psrecord(c_str_ptr(m_rootDir), m_flags, (*m_entriesBegin).get_find_data());
#endif /* 0 */

        if(NULL == m_current)
        {
            rc = RECLS_RC_OUT_OF_MEMORY;
        }
        else
        {
            rc = RECLS_RC_OK;
        }
    }
    else
    {
        if(m_directoriesBegin == m_directories.end())
        {
            rc = RECLS_RC_NO_MORE_DATA;
        }
        else
        {
            do
            {
#if 0
                // The way glob_sequence works
//              m_dnode = ReclsFtpSearchDirectoryNode::FindAndCreate(m_flags, m_rootDir, *m_directoriesBegin, m_pattern);
#else /* ? 0 */
                // The way basic_findfile_sequence<> works
                m_dnode = ReclsFtpSearchDirectoryNode::FindAndCreate(m_connection, m_flags, c_str_ptr(*m_directoriesBegin), c_str_ptr(m_pattern));
#endif /* 0 */

            } while(NULL == m_dnode && ++m_directoriesBegin != m_directories.end());

            rc = (NULL == m_dnode) ? RECLS_RC_NO_MORE_DATA : RECLS_RC_OK;
        }
    }

    if(RECLS_SUCCEEDED(rc))
    {
        recls_assert(is_valid_());
    }

    return rc;
}

recls_bool_t ReclsFtpSearchDirectoryNode::is_valid_() const
{
    function_scope_trace("ReclsFtpSearchDirectoryNode::is_valid_");

    recls_rc_t  rc  =   RECLS_RC_OK;

#if !defined(__STLSOFT_COMPILER_IS_MWERKS) && \
    !defined(__STLSOFT_COMPILER_IS_WATCOM)
# ifdef STLSOFT_CF_PRAGMA_MESSAGE_SUPPORT
#  pragma message("Flesh these out")
# endif /* STLSOFT_CF_PRAGMA_MESSAGE_SUPPORT */
    if(RECLS_SUCCEEDED(rc))
    {

    }

    if(RECLS_SUCCEEDED(rc))
    {
    }

    if(RECLS_SUCCEEDED(rc))
    {
    }

    if(RECLS_SUCCEEDED(rc))
    {
    }
#endif /* compiler */

    // (i) Either we are enumerating files (m_current != NULL) or directories (m_dnode != NULL), but not both
    recls_assert(NULL == m_current || NULL == m_dnode);
    // (ii) Either we are enumerating files (m_current != NULL) or there are no more files to enumerate
    recls_assert(NULL != m_current || m_entriesBegin == m_entries.end());

    return RECLS_SUCCEEDED(rc);
}

recls_rc_t ReclsFtpSearchDirectoryNode::GetNext()
{
    function_scope_trace("ReclsFtpSearchDirectoryNode::GetNext");

    recls_assert(is_valid_());

    /* Searching operates as follows:
     *
     * 1. Return all the contents of the files
     * 2. Return the contents of the directories.
     *
     * Hence, if m_dnode is non-NULL, we've already searched
     */

    /* States:
     *
     * - Iterating files: m_entriesBegin != m_entries.end(), m_dnode is NULL, m_directoriesBegin != m_directories.end()
     * - Iterating directories: m_directoriesBegin != m_directories.end(), m_dnode is non-NULL, m_current is NULL
     *
     */

    // Invariants

    // (i) Either we are enumerating files (m_current != NULL) or directories (m_dnode != NULL), but not both
    recls_assert(NULL == m_current || NULL == m_dnode);
    // (ii) Either we are enumerating files (m_current != NULL) or there are no more files to enumerate
    recls_assert(NULL != m_current || m_entriesBegin == m_entries.end());

    recls_rc_t  rc = RECLS_RC_NO_MORE_DATA;

    if(NULL != m_current)
    {
        // Currently enumerating through the files

        recls_assert(m_entriesBegin != m_entries.end());
        recls_assert(NULL == m_dnode);

        // Advance, and check for end of sequence
        ++m_entriesBegin;

        Entry_Release(m_current);
        if(m_entriesBegin != m_entries.end())
        {
            // Still enumerating, so just update m_current
            m_current = create_entryinfo_from_psrecord(c_str_ptr(m_rootDir), m_flags, (*m_entriesBegin).get_find_data());

            rc = RECLS_RC_OK;
        }
        else
        {
            // No more left in the files sequence, so delete m_current
            m_current = NULL;

            rc = RECLS_RC_NO_MORE_DATA;
        }
    }

    if(NULL == m_current)
    {
        // Now we are either enumerating the directories, or we've already done so

        if(NULL != m_dnode)
        {
            // Currently enumerating the directories
            rc = m_dnode->GetNext();

            if(RECLS_RC_NO_MORE_DATA == rc)
            {
                ++m_directoriesBegin;

                delete m_dnode;

                m_dnode = NULL;
            }
        }

        if(m_directoriesBegin == m_directories.end())
        {
            // Enumeration is complete.
            rc = RECLS_RC_NO_MORE_DATA;
        }
        else
        {
            if(NULL == m_dnode)
            {
                do
                {
                    // Creation of the node will cause it to enter the first enumeration
                    // state. However, if there are no matching

                    recls_assert(m_directoriesBegin != m_directories.end());

#if 0
                    // The way glob_sequence works
//                  m_dnode = ReclsFtpSearchDirectoryNode::FindAndCreate(m_flags, m_rootDir, *m_directoriesBegin, m_pattern);
#else /* ? 0 */
                    // The way basic_findfile_sequence<> works
                    m_dnode = ReclsFtpSearchDirectoryNode::FindAndCreate(m_connection, m_flags, c_str_ptr(*m_directoriesBegin), c_str_ptr(m_pattern));
#endif /* 0 */

                    if(NULL != m_dnode)
                    {
                        rc = RECLS_RC_OK;
                    }
                    else
                    {
                        ++m_directoriesBegin;
                    }

                } while(NULL == m_dnode && m_directoriesBegin != m_directories.end());
            }
        }
    }

    recls_assert(is_valid_());

    return rc;
}

recls_rc_t ReclsFtpSearchDirectoryNode::GetDetails(recls_info_t *pinfo)
{
    function_scope_trace("ReclsFtpSearchDirectoryNode::GetDetails");

    recls_assert(is_valid_());

    recls_rc_t  rc;

    recls_assert(NULL != pinfo);
    recls_assert(NULL == m_current || NULL == m_dnode);

    if(NULL != m_current)
    {
        // Currently searching for files from the current directory

        recls_assert(NULL == m_dnode);

        rc = Entry_Copy(m_current, pinfo);

#if defined(_DEBUG) && \
    defined(RECLS_PLATFORM_IS_WIN32)
        {
            size_t      cch =   Recls_GetPathProperty(m_current, NULL, 0);
            string_type buffer(cch, ' ');

            Recls_GetPathProperty(m_current, &buffer[0], 1 + buffer.length());

# if defined(RECLS_CHAR_TYPE_IS_CHAR)
            debug_printf("    [%s]\n", c_str_ptr(buffer));
# elif defined(RECLS_CHAR_TYPE_IS_WCHAR)
            debug_printf("    [%s]\n", static_cast<char const*>(winstl::w2a(buffer)));
# endif /* RECLS_CHAR_TYPE_IS_???? */
        }
#endif /* _DEBUG */
    }
    else if(NULL != m_dnode)
    {
        recls_assert(NULL == m_current);

        // Sub-directory searching is active, so get from there.

        rc = m_dnode->GetDetails(pinfo);
    }
    else
    {
        // Enumeration has completed
        rc = RECLS_RC_NO_MORE_DATA;
    }

    recls_assert(is_valid_());

    return rc;
}

recls_rc_t ReclsFtpSearchDirectoryNode::GetNextDetails(recls_info_t *pinfo)
{
    function_scope_trace("ReclsFtpSearchDirectoryNode::GetNextDetails");

    recls_assert(is_valid_());
    recls_assert(NULL != pinfo);

    recls_rc_t  rc  =   GetNext();

    if(RECLS_SUCCEEDED(rc))
    {
        rc = GetDetails(pinfo);
    }

    recls_assert(is_valid_());

    return rc;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */
