/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls_api_win32.cpp
 *
 * Purpose:     This file contains the Win32 versions of the recls API.
 *
 * Created:     16th August 2003
 * Updated:     6th March 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright 2003-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#define STLSOFT_STRING_ACCESS_NO_STD_STRING /* We don't care about std::string in recls */

#include "recls.h"
#include "recls_internal.h"
#include "recls_assert.h"
#include "recls_util.h"

#if !defined(RECLS_COMPILER_IS_MSVC) || \
    _MSC_VER >= 1200
# define USE_SEARCHSPEC_SEQ
#endif /* compiler */

#include <stlsoft.h>
#if !defined(RECLS_COMPILER_IS_VECTORC) && \
    (   !defined(RECLS_COMPILER_IS_MSVC) || \
        _MSC_VER != 1310)
# include <stlsoft_nulldef.h>
#endif /* compiler */
#ifdef USE_SEARCHSPEC_SEQ
# include <stlsoft_simple_string.h>
# include <stlsoft_string_tokeniser.h>
#endif /* USE_SEARCHSPEC_SEQ */

#include <winstl_file_path_buffer.h>
#include <winstl_filesystem_traits.h>
#if defined(RECLS_CHAR_TYPE_IS_WCHAR)
# include <winstl_char_conversions.h>
#endif /* RECLS_CHAR_TYPE_IS_???? */
#include <winstl_path.h>

#include <windows.h>
#include <ctype.h>

#include "recls_debug.h"

#include <algorithm>


// file system
#include "ReclsFileSearch.h"
#include "ReclsFileSearchDirectoryNode_win32.h"

// file entry
#include "EntryFunctions.h"

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

#ifdef USE_SEARCHSPEC_SEQ
stlsoft_ns_using(basic_simple_string)
stlsoft_ns_using(string_tokeniser)
#endif /* USE_SEARCHSPEC_SEQ */

typedef winstl_ns_qual(filesystem_traits)<recls_char_t>         traits_type; // We need to do this, because VC is a bit thick
typedef winstl_ns_qual(basic_file_path_buffer)<recls_char_t>    file_path_buffer_type;
typedef winstl_ns_qual(basic_path)<recls_char_t>                path_type;

stlsoft_ns_using(c_str_ptr)

/* /////////////////////////////////////////////////////////////////////////////
 * Global objects
 */

#if defined(_DEBUG) && \
    defined(RECLS_PLATFORM_IS_WIN32)

/* static */ tss_index  function_scope::sm_index;

#endif /* _DEBUG */

/* /////////////////////////////////////////////////////////////////////////////
 * Constants
 */

static const recls_char_t   s_defaultSearchRoot[]       =   RECLS_LITERAL(".");
static const recls_char_t   s_localDir[]                =   RECLS_LITERAL(".");
static const recls_char_t   s_parentDir[]               =   RECLS_LITERAL("..");

static const recls_char_t   s_protocolSep[]             =   RECLS_LITERAL("://");
static const recls_char_t   s_protocolFile[]            =   RECLS_LITERAL("file");
static const recls_char_t   s_protocolFTP[]             =   RECLS_LITERAL("ftp");
static const recls_char_t   s_protocolWin32Registry[]   =   RECLS_LITERAL("win/registry");

/* /////////////////////////////////////////////////////////////////////////////
 * Search control
 */

RECLS_FNDECL(recls_rc_t) Recls_Search(  recls_char_t const  *searchRoot
                                    ,   recls_char_t const  *pattern
                                    ,   recls_uint32_t      flags
                                    ,   hrecls_t            *phSrch)
{
    function_scope_trace("Recls_Search");

    recls_assert(NULL != phSrch);

    *phSrch = static_cast<hrecls_t>(0);

    // (1) Determine the protocol

    // An unspecified search is for the file-system
    SearchProtocol::SearchProtocol  protocol    =   SearchProtocol::fileSystem;
    recls_rc_t                      rc;

    // Determine the search protocol, and strip the protocol from the search root
    recls_char_t const  *sep;

    if( NULL != searchRoot &&
        NULL != (sep = traits_type::str_str(searchRoot, s_protocolSep)))
    {
        if(sep != searchRoot)
        {
            file_path_buffer_type   prot;

            traits_type::str_n_copy(&prot[0], searchRoot, prot.size());
            prot[sep - searchRoot] = '\0';

            if(0 == traits_type::str_compare(s_protocolFile, c_str_ptr(prot)))
            {
                protocol = SearchProtocol::fileSystem;
            }
            else if(0 == traits_type::str_compare(s_protocolFTP, c_str_ptr(prot)))
            {
                protocol = SearchProtocol::ftp;
            }
            else if(0 == traits_type::str_compare(s_protocolWin32Registry, c_str_ptr(prot)))
            {
                protocol = SearchProtocol::win32Registry;
            }
            else
            {
                protocol = SearchProtocol::unknown;
            }
        }

        searchRoot = sep + traits_type::str_len(s_protocolSep);
    }

    // (2) Default the input parameters

    // Default the search root
    if( NULL == searchRoot ||
        0 == *searchRoot)
    {
        searchRoot = s_defaultSearchRoot;
    }

    // Default the pattern
    if(NULL == pattern)
    {
        pattern = traits_type::pattern_all();
    }

    // Cater for the situation whereby the pattern is a file
    path_type               path(pattern);
    file_path_buffer_type   searchRoot_;

    if(path.exists())
    {
        path.make_absolute(true);
        path.canonicalise(true);

        recls_char_t const  *file   =   path.get_file();
        const size_t        cch     =   file - path.c_str();

        traits_type::str_n_copy(&searchRoot_[0], path.c_str(), cch);
        searchRoot_[cch] = '\0';
        searchRoot = searchRoot_.c_str();

        pattern = file;
    }

    // Default the flags
    if(0 == (flags & RECLS_F_TYPEMASK))
    {
        flags |= RECLS_F_FILES;
    }

    // (3) Validate the protocol & parameters

    // Since Win32 does not support all search types, we need to inform
    // the caller if they ask to create a search that can never be
    // satisfied.
    switch(protocol)
    {
        case    SearchProtocol::unknown:
            rc = RECLS_RC_INVALID_SEARCH_PROTOCOL;
            break;
        case    SearchProtocol::fileSystem:
            rc = RECLS_RC_OK;
            if(0 == (flags & (RECLS_F_FILES | RECLS_F_DIRECTORIES)))
            {
                rc = RECLS_RC_INVALID_SEARCH_TYPE;
            }
            // Validate the pattern.
            else if('\0' == *pattern)
            {
                rc = RECLS_RC_NO_MORE_DATA;
            }
            break;
        case    SearchProtocol::ftp:
        case    SearchProtocol::win32Registry:
            rc = RECLS_RC_NOT_IMPLEMENTED;
            break;
        default:
            rc = RECLS_RC_UNEXPECTED;
            break;
    }

    // Ensure that if '.' was specified as a/the pattern, it is not
    // for recursive searches.
# ifdef USE_SEARCHSPEC_SEQ
    // "." can only be specified as a pattern in non-recursive searches
    if(RECLS_SUCCEEDED(rc))
    {
        typedef basic_simple_string<recls_char_t>       string_t;
# if defined(RECLS_COMPILER_IS_WATCOM) || \
    (   defined(RECLS_COMPILER_IS_MSVC) && \
        _MSC_VER < 1300)
        typedef string_tokeniser<string_t, char>        tokeniser_t;

        tokeniser_t                 tokens(pattern, *Recls_GetPathSeparator());

        tokeniser_t::const_iterator begin   =   tokens.begin();
        tokeniser_t::const_iterator end     =   tokens.end();

        for(; begin != end; ++begin)
        {
            tokeniser_t::value_type const   value   =   *begin;

            if( value == s_parentDir ||
                (   (flags & RECLS_F_RECURSIVE) &&
                    value == s_localDir))
            {
                rc = RECLS_RC_DOT_RECURSIVE_SEARCH;
                break;
            }
        }

# else /* ? compiler */
        typedef string_tokeniser<string_t, string_t>    tokeniser_t;

        tokeniser_t     tokens(pattern, Recls_GetPathSeparator());

        if( tokens.end() != stlsoft_ns_qual_std(find)(tokens.begin(), tokens.end(), string_t(s_parentDir)) ||
            (   0 != (flags & RECLS_F_RECURSIVE) &&
                tokens.end() != stlsoft_ns_qual_std(find)(tokens.begin(), tokens.end(), string_t(s_localDir))))
        {
            rc = RECLS_RC_DOT_RECURSIVE_SEARCH;
        }
# endif /* compiler */
    }
#endif /* USE_SEARCHSPEC_SEQ */

    // (4) Dispatch to the appropriate implementor

    if(RECLS_SUCCEEDED(rc))
    {
        switch(protocol)
        {
            case    SearchProtocol::fileSystem:
                {
                    ReclsFileSearch *si;

                    rc = ReclsFileSearch::FindAndCreate(searchRoot, pattern, flags, &si);

                    if(RECLS_SUCCEEDED(rc))
                    {
                        *phSrch = ReclsSearch::ToHandle(si);

                        rc = RECLS_RC_OK;
                    }
                }
                break;
            case    SearchProtocol::ftp:
                rc = RECLS_RC_INVALID_SEARCH_PROTOCOL;
                break;
            case    SearchProtocol::win32Registry:
                rc = RECLS_RC_INVALID_SEARCH_PROTOCOL;
                break;
            default:
                rc = RECLS_RC_UNEXPECTED;
                break;
        }
    }

    recls_message_assert("Must not return a success code with a null search handle", RECLS_RC_OK != rc || NULL != *phSrch);

    return rc;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Property elicitation
 */

RECLS_FNDECL(size_t) Recls_GetDirectoryPathProperty(    recls_info_t    fileInfo
                                                    ,   recls_char_t    *buffer
                                                    ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetDirectoryPathProperty");

    recls_assert(NULL != fileInfo);

    struct recls_strptrs_t directoryPath =
    {
            fileInfo->path.begin    /* Directory path is defined by start of path ... */
        ,   fileInfo->directory.end /* ... to end of directory. */
    };

    return Recls_GetStringProperty_(&directoryPath, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetShortFileProperty(    recls_info_t    fileInfo
                                                ,   recls_char_t    *buffer
                                                ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetShortFileProperty");

    recls_assert(NULL != fileInfo);

    return Recls_GetStringProperty_(&fileInfo->shortFile, buffer, cchBuffer);
}

RECLS_FNDECL(void) Recls_GetDriveProperty(  recls_info_t    fileInfo
                                        ,   recls_char_t    *pchDrive)
{
    function_scope_trace("Recls_GetDriveProperty");

    recls_assert(NULL != fileInfo);
    recls_assert(NULL != pchDrive);

    // Because, as of version 1.5.1, this function can also be called for 
    // FTP files, which will not have a drive, we need to check for it,
    // and return a nul char if it's not a 

    *pchDrive = (':' == fileInfo->path.begin[1]) ? (recls_char_t)toupper(fileInfo->drive) : '\0';
}

RECLS_FNDECL(recls_bool_t) Recls_IsFileReadOnly(recls_info_t fileInfo)
{
    function_scope_trace("Recls_IsFileReadOnly");

    recls_assert(NULL != fileInfo);

    return fileInfo->attributes & FILE_ATTRIBUTE_READONLY;
}

RECLS_FNDECL(recls_bool_t) Recls_IsFileDirectory(recls_info_t fileInfo)
{
    function_scope_trace("Recls_IsFileDirectory");

    recls_assert(NULL != fileInfo);

    return fileInfo->attributes & FILE_ATTRIBUTE_DIRECTORY;
}

RECLS_FNDECL(recls_bool_t) Recls_IsFileLink(recls_info_t fileInfo)
{
    function_scope_trace("Recls_IsFileLink");

    recls_assert(NULL != fileInfo);
    ((void)fileInfo);

    return false;
}

RECLS_FNDECL(recls_time_t) Recls_GetCreationTime(recls_info_t fileInfo)
{
    function_scope_trace("Recls_GetCreationTime");

    recls_assert(NULL != fileInfo);

    return fileInfo->creationTime;
}

RECLS_FNDECL(recls_time_t) Recls_GetLastStatusChangeTime(recls_info_t fileInfo)
{
    function_scope_trace("Recls_GetLastStatusChangeTime");

    recls_assert(NULL != fileInfo);

    return fileInfo->modificationTime;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */
