/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls_fileinfo_win32.cpp
 *
 * Purpose:     Win32 implementation for the file information blocks of the recls API.
 *
 * Created:     16th August 2003
 * Updated:     10th January 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright 2003-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include "recls.h"
#include "recls_internal.h"
#include "recls_assert.h"

#include <stlsoft.h>
#if !defined(RECLS_COMPILER_IS_VECTORC)
# include <stlsoft_nulldef.h>
#endif /* compiler */

#include <winstl.h>
#include <winstl_atomic_functions.h>

#include "EntryFunctions.h"

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

winstl_ns_using(ws_sint32_t)
winstl_ns_using(atomic_preincrement)
winstl_ns_using(atomic_predecrement)

/* /////////////////////////////////////////////////////////////////////////////
 * Constants and definitions
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace
{
#else
static
#endif /* !RECLS_NO_NAMESPACE */

volatile recls_sint32_t s_createdInfoBlocks =   0;
volatile recls_sint32_t s_sharedInfoBlocks  =   0;

#if !defined(RECLS_NO_NAMESPACE)
} // namespace recls
#endif /* !RECLS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Typedefs
 */

struct counted_recls_info_t
{
    volatile recls_sint32_t rc;
    recls_uint32_t          _;
    struct recls_fileinfo_t info;
};

/* /////////////////////////////////////////////////////////////////////////////
 * Helpers
 */

inline recls_sint32_t RC_PreIncrement(recls_sint32_t volatile *p)
{
    return atomic_preincrement(static_cast<ws_sint32_t volatile*>(static_cast<void volatile*>(p)));
}

inline recls_sint32_t RC_PreDecrement(recls_sint32_t volatile *p)
{
    return atomic_predecrement(static_cast<ws_sint32_t volatile*>(static_cast<void volatile*>(p)));
}

inline struct counted_recls_info_t *counted_info_from_info(recls_info_t i)
{
    recls_assert(i != NULL);

    // can't be bothered with all the C++ casts here!
    return (struct counted_recls_info_t *)((recls_byte_t*)i - offsetof(counted_recls_info_t, info));
}

inline recls_info_t info_from_counted_info(struct counted_recls_info_t * ci)
{
    recls_assert(ci != NULL);

    return &ci->info;
}

/* /////////////////////////////////////////////////////////////////////////////
 * File info functions
 */

RECLS_FNDECL(recls_info_t) Entry_Allocate(size_t cb)
{
    // Simply allocate a lock-count prior to the main memory (but do it on an 8-byte block)
    counted_recls_info_t    *ci     =   static_cast<counted_recls_info_t*>(malloc(cb - sizeof(struct recls_fileinfo_t) + sizeof(struct counted_recls_info_t)));
    recls_info_t            info;

    if(NULL == ci)
    {
        info = NULL;
    }
    else
    {
        ci->rc  =   1; // One initial reference
        info    =   info_from_counted_info(ci);

        RC_PreIncrement(&s_createdInfoBlocks);
    }

    return info;
}

RECLS_FNDECL(void) Entry_Release(recls_info_t fileInfo)
{
    if(NULL != fileInfo)
    {
        counted_recls_info_t    *pci    =   counted_info_from_info(fileInfo);

        if(0 == RC_PreDecrement(&pci->rc))
        {
            free(pci);

            RC_PreDecrement(&s_createdInfoBlocks);
        }
        else
        {
            RC_PreDecrement(&s_sharedInfoBlocks);
        }
    }
}

RECLS_FNDECL(recls_rc_t) Entry_Copy(recls_info_t fileInfo, recls_info_t *pinfo)
{
    recls_assert(NULL != pinfo);

    if(NULL != fileInfo)
    {
        counted_recls_info_t    *pci    =   counted_info_from_info(fileInfo);

        RC_PreIncrement(&pci->rc);
        RC_PreIncrement(&s_sharedInfoBlocks);
    }

    *pinfo = fileInfo;

    return RECLS_RC_OK;
}

RECLS_FNDECL(void) Entry_BlockCount(recls_sint32_t *pcCreated, recls_sint32_t *pcShared)
{
    recls_assert(NULL != pcCreated);
    recls_assert(NULL != pcShared);

    // Because on 3.51 and 95, the InterlockedInc/Decrement functions do not 
    // return the values, we're going to fudge it

    RC_PreIncrement(&s_createdInfoBlocks);
    recls_sint32_t createdInfoBlocks    =   s_createdInfoBlocks;
    RC_PreDecrement(&s_createdInfoBlocks);
    *pcCreated                          =   createdInfoBlocks - 1;

    RC_PreIncrement(&s_sharedInfoBlocks);
    recls_sint32_t sharedInfoBlocks     =   s_sharedInfoBlocks;
    RC_PreDecrement(&s_sharedInfoBlocks);
    *pcShared                           =   sharedInfoBlocks - 1;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */
