/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls_ftp_api_win32.cpp
 *
 * Purpose:     This file contains the Win32 versions of the FTP part of the recls API.
 *
 * Created:     15th August 2004
 * Updated:     6th March 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright 2004-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#define STLSOFT_STRING_ACCESS_NO_STD_STRING /* We don't care about std::string in recls */

#include "recls.h"
#include "recls_internal.h"
#include "recls_assert.h"
#include "recls_util.h"
#include "recls_ftp.h"

#if !defined(RECLS_COMPILER_IS_MSVC) || \
    _MSC_VER >= 1200
# define USE_SEARCHSPEC_SEQ
#endif /* compiler */

#include <stlsoft.h>
#if !defined(RECLS_COMPILER_IS_VECTORC) && \
    (   !defined(RECLS_COMPILER_IS_MSVC) || \
        _MSC_VER != 1310)
# include <stlsoft_nulldef.h>
#endif /* compiler */
#ifdef USE_SEARCHSPEC_SEQ
# include <stlsoft_simple_string.h>
# include <stlsoft_string_tokeniser.h>
#endif /* USE_SEARCHSPEC_SEQ */

#include <winstl_file_path_buffer.h>
#include <winstl_filesystem_traits.h>
#if defined(RECLS_CHAR_TYPE_IS_WCHAR)
# include <winstl_char_conversions.h>
#endif /* RECLS_CHAR_TYPE_IS_???? */

#include <windows.h>
#include <ctype.h>

#include "recls_debug.h"

#include <algorithm>

#if defined(RECLS_DELAY_LOAD_WININET)
# ifdef INETSTL_INCL_H_INETSTL
#  error INETSTL_INCL_H_INETSTL
# endif /* INETSTL_INCL_H_INETSTL */
# include "recls_wininet_dl.h"
#endif /* RECLS_DELAY_LOAD_WININET */

// ftp
#include "ReclsFtpSearch.h"

// win32Registry

#include "EntryFunctions.h"

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

#ifdef USE_SEARCHSPEC_SEQ
stlsoft_ns_using(basic_simple_string)
stlsoft_ns_using(string_tokeniser)
#endif /* USE_SEARCHSPEC_SEQ */

typedef winstl_ns_qual(filesystem_traits)<recls_char_t>         traits_type; // We need to do this, because VC is a bit thick
typedef winstl_ns_qual(basic_file_path_buffer)<recls_char_t>    file_path_buffer_type;

stlsoft_ns_using(c_str_ptr)

/* /////////////////////////////////////////////////////////////////////////////
 * Constants
 */


static const recls_char_t   s_localDir[]                =   RECLS_LITERAL(".");
static const recls_char_t   s_parentDir[]               =   RECLS_LITERAL("..");

static const recls_char_t   s_protocolSep[]             =   RECLS_LITERAL("://");
static const recls_char_t   s_protocolFile[]            =   RECLS_LITERAL("file");
static const recls_char_t   s_protocolFTP[]             =   RECLS_LITERAL("ftp");
static const recls_char_t   s_protocolWin32Registry[]   =   RECLS_LITERAL("win/registry");

/* /////////////////////////////////////////////////////////////////////////////
 * Search control
 */

RECLS_FNDECL(recls_rc_t) Recls_SearchFtp(   recls_char_t const  *host
                                    ,       recls_char_t const  *username
                                    ,       recls_char_t const  *password
                                    ,       recls_char_t const  *searchRoot
                                    ,       recls_char_t const  *pattern
                                    ,       recls_uint32_t      flags
                                    ,       hrecls_t            *phSrch)
{
    function_scope_trace("Recls_SearchFtp");

    recls_assert(NULL != phSrch);

    *phSrch = static_cast<hrecls_t>(0);

    // (1) Validate that the protocol is either unspecified, or is FTP

    // An unspecified search is for the FTP
    SearchProtocol::SearchProtocol  protocol    =   SearchProtocol::ftp;
    recls_rc_t                      rc          =   RECLS_RC_OK;

    // Determine the search protocol, and strip the protocol from the search root
    recls_char_t const  *sep;

    if( NULL != searchRoot &&
        NULL != (sep = traits_type::str_str(searchRoot, s_protocolSep)))
    {
        if(sep != searchRoot)
        {
            file_path_buffer_type   prot;

            traits_type::str_n_copy(&prot[0], searchRoot, prot.size());
            prot[sep - searchRoot] = '\0';

            if(0 == traits_type::str_compare(s_protocolFile, c_str_ptr(prot)))
            {
                protocol = SearchProtocol::fileSystem;
            }
            else if(0 == traits_type::str_compare(s_protocolFTP, c_str_ptr(prot)))
            {
                protocol = SearchProtocol::ftp;
            }
            else if(0 == traits_type::str_compare(s_protocolWin32Registry, c_str_ptr(prot)))
            {
                protocol = SearchProtocol::win32Registry;
            }
            else
            {
                protocol = SearchProtocol::unknown;
            }
        }

        searchRoot = sep + traits_type::str_len(s_protocolSep);
    }

    if(protocol != SearchProtocol::ftp)
    {
        rc = RECLS_RC_INVALID_SEARCH_PROTOCOL;
    }
    else
    {
        // (2) Default the input parameters

        // Default the search root
        if( NULL == searchRoot ||
            0 == *searchRoot)
        {
            searchRoot = RECLS_LITERAL("/"); // FTP always rooted at ./.
        }

        // Default the pattern
        if(NULL == pattern)
        {
            pattern = RECLS_LITERAL("*"); // FTP always uses '*' as wildcard
        }

        // Default the flags
        if(0 == (flags & RECLS_F_TYPEMASK))
        {
            flags |= RECLS_F_FILES;
        }

        // (3) Validate the protocol & parameters

        // Ensure that if '.' was specified as a/the pattern, it is not
        // for recursive searches.
# ifdef USE_SEARCHSPEC_SEQ
        // "." can only be specified as a pattern in non-recursive searches
        if(RECLS_SUCCEEDED(rc))
        {
            typedef basic_simple_string<recls_char_t>       string_t;
# if defined(RECLS_COMPILER_IS_WATCOM) || \
        (   defined(RECLS_COMPILER_IS_MSVC) && \
            _MSC_VER < 1300)
            typedef string_tokeniser<string_t, char>        tokeniser_t;

            tokeniser_t                 tokens(pattern, *Recls_GetPathSeparator());

            tokeniser_t::const_iterator begin   =   tokens.begin();
            tokeniser_t::const_iterator end     =   tokens.end();

            for(; begin != end; ++begin)
            {
                tokeniser_t::value_type const   value   =   *begin;

                if( value == s_parentDir ||
                    (   /* (flags & RECLS_F_RECURSIVE) &&
                         */value == s_localDir))
                {
                    rc = RECLS_RC_DOT_RECURSIVE_SEARCH;
                    break;
                }
            }

# else /* ? compiler */
            typedef string_tokeniser<string_t, string_t>    tokeniser_t;

            tokeniser_t     tokens(pattern, Recls_GetPathSeparator());

            if( tokens.end() != stlsoft_ns_qual_std(find)(tokens.begin(), tokens.end(), string_t(s_parentDir)) ||
                (   0 != (flags & RECLS_F_RECURSIVE) &&
                    tokens.end() != stlsoft_ns_qual_std(find)(tokens.begin(), tokens.end(), string_t(s_localDir))))
            {
                rc = RECLS_RC_DOT_RECURSIVE_SEARCH;
            }
# endif /* compiler */
        }
#endif /* USE_SEARCHSPEC_SEQ */

        // (4) Dispatch to the appropriate implementor

        if(RECLS_SUCCEEDED(rc))
        {
            switch(protocol)
            {
                case    SearchProtocol::ftp:
                    {
                        ReclsFtpSearch *si;

                        rc = ReclsFtpSearch::FindAndCreate(host, username, password, searchRoot, pattern, flags, &si);

                        if(RECLS_SUCCEEDED(rc))
                        {
                            *phSrch = ReclsSearch::ToHandle(si);

                            rc = RECLS_RC_OK;
                        }
                    }
                    break;
                default:
                    recls_assert(0);
                    rc = RECLS_RC_UNEXPECTED;
                    break;
            }
        }
    }

    return rc;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */
