/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 + Coded by Larry Cowan
 + March 24, 2004
 + (ported from file wildmatch.d version:0.03)
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 +
 + This class stores a unix-style wildcard pattern in an instance and then
 + can test filenames for matching the pattern. * (zero or more of any
 + character) and ? (any one character) and [abc] and [^xyz] are supported
 + in wildcard content, as is [ac-xzBY] and some such.
 + This matching is case-insensitive for win32 and case-sensitive for any
 + other platform - or it may be specified to the constructor.
 +
 + Usage:   void initpattern(char *pattern, bool ignorecase)
 +          bool test(filename1)
 +          bool test(filename2)
 +
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 +  Copyright 2004 by Larry F. Cowan
 +  Distribution rights:  This code may be used for any purpose whatever,
 +      commercial or otherwise, without specific permission from the 
 +      author(s), as long as they are not misrepresented in any way, the
 +      author name(s) are kept in the source, and this notice is included.
 +      Changes should indicate at least an author name and date.  No listed
 +      author takes on any responsibility or liability for your use of this
 +      module. It's free.  Caveat emptor!  Any underlying (imported) modules
 +      are under their own licensing and this is not affected here. 
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include <unistd.h>
#include <assert.h>

#define bool	int
#define true	1
#define false	0

#	define Wildmatch
#	include <string.h>
#	include <stdio.h>

/* Globals, so that pattern can be set and reused for lots of filenames */
char* pattern = NULL;
bool case_insensitive;

/* Since D has an overloaded version like this */
/* (almost, it only copies the string if changed) */
char *toupper__ ( char *str )
{
    char *dup, *d;

    dup = (char*)malloc(1+strlen(str));
    strcpy(dup,str);
    for (d=dup; *d ;d++)
        if (*d >= 'a' && *d <= 'z')
            *d += 'A' - 'a';
    return dup;
}

/* Save the pattern (in place of any previously given one) */
void initpattern ( char* initpattern, bool ignorecase )
{
    char *p;

    if (pattern != NULL) { /* free previous pattern, if any */
        free(pattern);
	pattern = NULL;
    }
    case_insensitive = ignorecase;
    if (initpattern && strlen(initpattern) > 0) {
        if (case_insensitive)
	    pattern = toupper__(initpattern); /* mallocs */
	else {
            pattern = (char *)malloc(1+strlen(initpattern));
            strcpy(pattern,initpattern);
	}
    } else {
        pattern = (char *)malloc(2);
        strcpy(pattern,"*");  /* default is accept everything */
    }
#ifdef wildmatch
    printf("Init(%d,%s): \"%s\"\n"
           ,strlen(pattern)
	   ,(case_insensitive?"true":"false")
	   ,pattern);
#endif
}

#ifdef wildmatch
void trace ( int i, char c, int q, int *pstates )
{
    if (q < strlen(pattern))
        printf("%d%c. setting pattern%d[%c] %s\n",i,c,q,pattern[q],(pstates[q]<1)?"off":"on");
    else
        printf("%d%c. setting pattern%d[plen] %s\n",i,c,q,(pstates[q]<1)?"off":"on");
}
#define TRACE(a,b,c) trace(a,b,c,pstates);
#else
#define TRACE(a,b,c)
#endif

bool test ( char* filename )
{
    int* pstates;
    bool caret;
    bool tempok;
    int plen;
    int i, j;
    char ch;
    char *f;

    assert( strlen(filename) > 0 );
    if (case_insensitive)
	filename = toupper__(filename);
    plen = strlen(pattern);
#ifdef wildmatch
    printf("\nPattern(%d): \"%s\" vs \"%s\"\n",plen,pattern,filename);
#endif
    pstates = (int*)malloc(sizeof(int) * (1+strlen(pattern)));
    for (i=0; i <= plen ;i++)
        pstates[i] = 0;
    if ( pattern[i=0] == '*' ) {
        for (j=1; j < plen ;j++)
            if ( pattern[j] != '*' ) /* scan successive stars */
                 break;
        if ( j == plen - 1 ) {
            if (case_insensitive)
	        free(filename);
            return true;   /* short-circuit = pattern is all stars */
	}
        pstates[j - 1] = 1; /* turn on last of group of stars */
        TRACE(i,'a',j-1)
        pstates[j] = 1;    /* accept - turn on following character */
        TRACE(i,'a',j)
    } else {
        pstates[i=0] = 1;
        TRACE(i,'a',0)
    }
    for (f=filename; (ch = *f) ;f++ ) {
#ifdef wildmatch
        printf("[%c]:\n",ch);
#endif
        pstates[plen] = 0;
        TRACE(i,'b',plen)
        for (i=plen-1; i >= 0 ;i--) {
            if ( pstates[i] == 0 ) /* if state inactive, ignore it */
                continue;
            switch ( pattern[i] ) {
              case '*':
                if ( i == plen - 1 ) {
                    if (case_insensitive)
	                free(filename);
                    return true;   /* short-circuit = pattern ends in '*' */
		}
                for (j=i+1; j < plen ;j++)
                    if ( pattern[j] != '*' ) /* scan through successive stars */
                         break;
                pstates[i] = 0;     /* turn off current star */
                TRACE(i,'c',i)
                pstates[j - 1] = 1; /* turn on last of group of stars */
                TRACE(i,'d',j-1)
                pstates[j] = 1;    /* accept - turn on following character */
                TRACE(i,'e',j)
                break;
              case '?':
                pstates[i] = 0;    /* accept - move to next char */
                TRACE(i,'f',i)
                pstates[i+1] = 1;  /* */
                TRACE(i,'g',i+1)
                break;
              case ']':
                if (case_insensitive)
	            free(filename);
                return false;  /* short-circuit - missing ']' */
              case '[':
                j = i;
                if (pattern[j+1] == '^') {
                    caret = true;
                    j++;
                } else
                    caret = false;
                for (tempok=caret,j++; pattern[j] != ']' ;j++) {
                    if ( j == plen ) {
                        if (case_insensitive)
	                    free(filename );
                        return false;  /* short-circuit - unclosed '[' */
		    }
 		    if ( (pattern[j+1] == '-'                 /* e.g., A-Z */
		       && (j += 2)  /* scan ahead, hit or miss */
                       && pattern[j-2] <= ch 
                       && pattern[j] >= ch)
      /* possible strangeness here if bad pattern like X-Q and ch is a Q */
      /* but I don't think I care, Z-a is even stranger in win32 (== A only) */
                      || pattern[j] == ch ) { /* or normal chr */
                        if ( caret )
                            tempok = false; /* matched in [^...] */
                        else
                            tempok = true;  /* matched in [...] */
		    }
		    /* else try the rest until ] */
                }
                if ( tempok ) {
                    pstates[i] = 0;  /* accept, go to next char in pattern */
                    TRACE(i,'h',i)
                    pstates[j+1] = 1;
                    TRACE(i,'i',j+1)
                } else {
                    pstates[i] = 0; /* [] failed */
                    TRACE(i,'j',i)
                }
                break;
              default:
                if ( ch == pattern[i] ) {
                    pstates[i] = 0;  /* accept, move to next char in patt */
                    TRACE(i,'k',i)
                    pstates[i+1] = 1;
                    TRACE(i,'l',i+1)
                } else {
                    pstates[i] = 0;  /* ch failed */
                    TRACE(i,'m',i)
                }
                break;
            }
        }
        for (i=plen-1; i >= 0 ;i--) { /* post-process any active stars */
            if ( pstates[i] == 0      /* in case we're done with filename */
              || pattern[i] != '*' )
                continue;
            for (j=i+1; j < plen ;j++)
                if ( pattern[j] != '*' ) /* scan through successive stars */
                     break;
            pstates[j] = 1;    /* turn on following character */
            TRACE(i,'n',j)
        }
    }
    if ( pstates[plen] > 0 ) {
#ifdef wildmatch
        printf("TRUE\n");
#endif
        if (case_insensitive)
            free(filename);
        return true;  /* end of pattern coincides with end of filename */
    }
#ifdef wildmatch
    printf("FALSE\n");
#endif
    if (case_insensitive)
        free(filename);
    return false;     /* else failed */
}

#ifdef Wildmatch
/* ****************************************** *
 *       main() - used for testing only       *
 * ****************************************** */

#if version == win32
#define Initpattern(x) initpattern(x,true)
#else
#define Initpattern(x) initpattern(x,false)
#endif

#define TESTnPRINT(n,cics) printf("wc%d%s(%s) is %s\n"       \
		               ,n                            \
		               ,cics ? "ci" : "cs"           \
			       ,w()                          \
			       ,test(fn) ? "true" : "false")

#define TESTPATTERN(n,xxx) Initpattern(xxx);                 \
		           TESTnPRINT(n,case_insensitive)

#define TESTpattern(n,xxx,tf) initpattern(xxx,tf);           \
		              TESTnPRINT(n,tf)

char* w(void) { return pattern; }

/* test user filename vs. test pattern list */
int main ( int argc, char *argv[] )
{
    char* fn = argv[1];

    if (argc != 2 && argc != 3) {
        printf("Usage: filename [pattern]\n");
        exit(0);
    }
    if (argc == 3) {          /* if a test pattern is given, use it (only) */
        TESTPATTERN(0,argv[2]);  /* at auto (compiled in) case_sensitivity */
        exit(0);
    }
    TESTpattern(1,"wildmatch.d",true); /* forcing 1 case-insensitive test */
    TESTpattern(2,"WILDMATCH.d",false); /* forcing 2 case-sensitive tests */
    TESTpattern(3,"*match.d",false);
    TESTPATTERN(4,"*");
    TESTPATTERN(5,"*.*");
    TESTPATTERN(6,"*.d");
    TESTPATTERN(7,"s*");
    TESTPATTERN(8,"S*.*");
    TESTPATTERN(9,"s*.d");
    TESTPATTERN(10,"?*");
    TESTPATTERN(11,"*?");
    TESTPATTERN(12,"**");
    TESTPATTERN(13,"??");
    TESTPATTERN(14,"*********");
    TESTPATTERN(15,"*?**?***?****?*****");
    TESTPATTERN(16,"wild[j-n][aaaaaeiou]tch.[dc]");
    TESTPATTERN(17,"WILD[^NJ-KL][^EIOU]TCH.[^D]XE");
    TESTPATTERN(18,"wildMATCH.D");
    TESTPATTERN(19,"wildmatch.d*");
    TESTPATTERN(20,"w?ld??*.?");
    TESTpattern(21,"a-z",false); /* force...cs */
    TESTpattern(22,"[b-y4-8]",true); /* force...ci */
    TESTnPRINT(23,true); /* repeat with same pattern */
    exit(0);
}
#endif
