################################################################################
# File:		InstallToWeb.rb
#
# Purpose:	Various utility functions for handling dates and versions
#
# Imports:  openrj, recls, synsoft/cmdline, synsoft/srcutil, UNPrompt
#
# Created:	6th March 2005
# Updated:  6th March 2005
#
# Home:		http://synesis.com.au/software
#
# Copyright 2005, Matthew Wilson and Synesis Software
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# - Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
#   any contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
################################################################################


################################################################################
# Version Info

SYNSOFT_VER_RB_INSTALLTOWEB_MAJOR		=	1
SYNSOFT_VER_RB_INSTALLTOWEB_MINOR		=	0
SYNSOFT_VER_RB_INSTALLTOWEB_REVISION	=	1
SYNSOFT_VER_RB_INSTALLTOWEB_EDIT		=	1

################################################################################
# Imports

require 'openrj'
require 'recls'
require 'synsoft/cmdline'
require 'synsoft/srcutil'
#require 'UNPrompt'

#abort if not puts Recls::usage()
#abort if not puts UNPrompt::usage()

# This script does the following tasks:

# 0. Check that the RECLS_VER environment variable exists, or has been passed in on the command-line
# 1. Check that the zip & md5 files exist
# 2. Check that the software/index.html files exists and is writeable
# 3. Check if the section for the current version does not exist in the HTML
# 4. Check that the files do not exist in the download directory
# 5. Check that the files do not exist in the directory 'online'
# 6. Write the section into the file
# 7. Copies the files into the destination

################################################################################
# Classes

class Info

		def initialize(fileName, fileExt, source)
				@fileName		=	fileName
				@fileExt		=	fileExt
				@source			=	source
				@destination	=	nil
				@bUpToDate		=	nil
		end

		def relative(dir)

				if dir.length
						last = dir[(dir.length) -1 ... dir.length]
						case	last
						when "\\", "/"
						else
								raise "Info::relative(): directory was not terminated with a path separator; dir: #{dir}"
						end
				end

				return dir + fileName + '.' + fileExt
				
		end

		attr_reader :fileName, :fileExt, :bUpToDate, :source, :destination
		attr_writer :bUpToDate, :destination

end # class Info
		
################################################################################
# Functions

def usage(message = nil)

		puts
		puts "USAGE: InstallToWeb[.rb] {-E<var0>=<val0> ... -E<varN>=<valN>} -D<deploy-file> -R<release-file>"
		puts
		if message
				puts "Error: #{message}"
		end

		exit

end # usage()

		
################################################################################
# Command-line processing

arguments 	= 	CmdLine::Arguments::new(ARGV)

values		=	arguments.values
options		=	arguments.options

################################################################################
# Constants

PROJECT_NAME		=	'recls'
PROJECT_VER_NAME	=	'RECLS_VER'

################################################################################
# Main processing

bReportOnly			=	nil
bVerbose			=	nil

deployFile			=	nil
releaseFile			=	nil

variables			=	{}

projVer				=	nil

options.each \
{ |option|
		case	option
		when "?"
				usage()
		when "v"
				bVerbose = 1
		else
				prefix 	= 	option[0 ... 1]
				value	=	option[1 ... option.length]
				case	prefix
				when	"E"
						if value =~ /(\w+)=(.*?)$/
								variables[$1] = $2
						else
								usage("Invalid option \"-E#{value}\"; should be \"-E<env-var-name>=<env-var-value\"")
						end
				when	"D"
						deployFile	=	value
				when	"R"
						releaseFile	=	value
				else
						usage("Unrecognised option: -#{option}")
				end
		end
}

if not deployFile
		usage("Must specify the deployment file")
end
if not releaseFile
		usage("Must specify the release file")
end

# 1. Open the configuration files

if bVerbose
		puts "  Opening the deployment file: #{deployFile}"
end
deployDb	=	OpenRJ::FileDatabase::new(deployFile, OpenRJ::ELIDE_BLANK_RECORDS)

deploy		=	deployDb[0]

if bVerbose
		puts "  Opening the release file:    #{releaseFile}"
end
releaseDb	=	OpenRJ::FileDatabase::new(releaseFile, OpenRJ::ELIDE_BLANK_RECORDS)

release		=	releaseDb[0]


# 2. Get the RECLS_VER environment variable, from one of:
# 
# - command-line
# - release file
# - environment

if not projVer and variables.has_key?(PROJECT_VER_NAME)
		projVer = variables[PROJECT_VER_NAME]
end

if not projVer and releaseDb[0].include?(PROJECT_VER_NAME)
		projVer = releaseDb[0][PROJECT_VER_NAME]
end

if not projVer and ENV.has_key?(PROJECT_VER_NAME)
		projVer = ENV[PROJECT_VER_NAME]
end

if not projVer
		usage("#{PROJECT_VER_NAME} not specified in command-line, or in release file, or in environment")
else
		projLabel = PROJECT_NAME + "_" + projVer.to_s.gsub(/\./, '_')
		if bVerbose
				puts "  Processing deployment for:   #{PROJECT_NAME}, version #{projVer}"
		end
end

INSERTIONPOINT_BEGIN		=	"#{PROJECT_NAME}-InsertionPoint-Begin"
INSERTIONPOINT_END			=	"#{PROJECT_NAME}-InsertionPoint-End"
INSERTIONPOINT_BEGIN_VER	=	"#{PROJECT_NAME}-InsertionPoint-Begin:#{projVer}"
INSERTIONPOINT_END_VER		=	"#{PROJECT_NAME}-InsertionPoint-End:#{projVer}"

# 3. Check that the ZIP & MD5 files exist

if bVerbose
		puts
		puts "  Verifying the existence of the .zip and .md5 files"
end

baseName = PROJECT_NAME + '-' + String(projVer)

#puts "baseName: #{baseName}"

fs = Recls::FileSearch::new(deploy['ReadyBinariesDir'], baseName + ".*", Recls::FILES)

feZip	=	fs.select { |fe| fe.fileExt == "zip" }
feMD5	=	fs.select { |fe| fe.fileExt == "md5" }

[[feZip, baseName + ".zip"], [feMD5, baseName + ".md5"] ].each { |pair| usage("The source file \"#{pair[1]}\" does not exist - aborting") if not pair[0][0] }

feZip = feZip[0]
feMD5 = feMD5[0]

if bVerbose
		puts "    Zip file: " + String(feZip) + " - exists"
		puts "    MD5 file: " + String(feMD5) + " - exists"
end

infos = [ Info::new(baseName, 'zip', feZip), Info::new(baseName, 'md5', feMD5) ]

# 4. Check that the software/index.html files exists and is writeable

if bVerbose
		puts
		puts "  Verifying the existence of the HTML files, and that they are writable"
end

indexPage		=	deploy['WebPage-Index']
downloadsPage	=	deploy['WebPage-Downloads']

if not File::exist?(indexPage)
		usage("Index page \"#{indexPage}\" does not exist")
elsif not File::stat(indexPage).writable?()
		puts "Index page \"#{indexPage}\" exists but is not writeable"
#		puts "Attempting to check it out from SourceSafe"
elsif bVerbose
		puts "    Index page:     " + indexPage
end

if not File::exist?(downloadsPage)
		usage("Downloads page \"#{downloadsPage}\" does not exist")
elsif not File::stat(downloadsPage).writable?()
		puts "Downloads page \"#{downloadsPage}\" exists but is not writeable"
#		puts "Attempting to check it out from SourceSafe"
elsif bVerbose
		puts "    Downloads page: " + downloadsPage
end

# 5. Check that the files do not exist in the download directory

if bVerbose
		puts
		puts "  Detecting presence of files in the upload directory"
end

webUploadDir	=	deploy['WebUploadDir']
if webUploadDir.length > 0 and webUploadDir[webUploadDir.length - 1] != '\\'
		webUploadDir = webUploadDir + "\\"
end

if bVerbose
		puts "    Upload directory: #{webUploadDir}"
end

fs = Recls::FileSearch::new(webUploadDir, baseName + ".*", Recls::FILES)

feZipDest	=	fs.select { |fe| fe.fileExt.downcase == "zip" }
feMD5Dest	=	fs.select { |fe| fe.fileExt.downcase == "md5" }

[[feZipDest, "zip"], [feMD5Dest, "md5"] ].each \
{ |pair|
		fe = pair[0][0]
		if fe and fe.fileExt != pair[1]
				usage("The destination file \"#{fe}\" has invalid case - aborting")
		end
}

#[[feZipDest, baseName + ".zip"], [feMD5Dest, baseName + ".md5"] ].each { |pair| puts "The destination file \"#{webUploadDir}\\#{pair[1]}\" does not exist" if not pair[0][0] }

feZipDest = feZipDest[0] if feZipDest
feMD5Dest = feMD5Dest[0] if feMD5Dest

infos[0].destination = feZipDest
infos[1].destination = feMD5Dest

#[ [feZip, feZipDest], [feMD5, feMD5Dest] ].each_with_index \
infos.each \
{ |info, i|
		src		=	info.source
		dest	=	info.destination
		if dest
				if src.modificationTime == dest.modificationTime
						if bVerbose
								puts "    The destination \"#{dest}\" is up to date"
						end
						info.bUpToDate = 1
				else
						if dest.isReadOnly
								usage("    The destination \"#{dest}\" is read-only - aborting!")
						elsif src.modificationTime < dest.modificationTime
								usage("    The destination \"#{dest}\" is NEWER than the source \"#{src}\" - aborting!")
						elsif src.modificationTime > dest.modificationTime
								if bVerbose
										puts "    " + webUploadDir + src.file + " exists, but is out of date"
								end
#						elsif src.size != dest.size
#								usage("    Different sizes; src: #{src.size}; dest: #{dest.size}")
						else
								if bVerbose
										puts "    #{dest} exists"
								end
						end
				end
		else
				if bVerbose
						puts "    " + webUploadDir + src.file + " does not exist"
				end
		end
}

# 6. Check that the files do not exist in the directory 'online'

if bVerbose
		puts
		puts "  Detecting presence of files in the online directory"
end

puts "    Not current implemented"

#host		=	deploy['FTPHost']
#userName	=	deploy['FTPUserName']

#prompt = UNPrompt::Prompt::new("FTP Access", "Enter Username / Password for [#{host}: #{userName}]", UNPrompt::PASSWORD)

#fs = Recls::FtpSearch::new(host, userName, prompt.password, deploy['FTPDownloadDir'], '#{baseName}.zip;#{baseName}.md5', Recls::FILES)
#
#fs.each \
#{ |fe|
#		case fe.fileExt
#		when "md5", "zip"
#				puts "Found: #{fe}"
#		end
#}

# 7. Copies the distribution files into the destination

if bVerbose
		puts
		puts "  Copying the distribution files into the upload directory"
end

infos.each \
{ |info|
		if info.bUpToDate
				if bVerbose
						puts "    #{info.relative(webUploadDir)} up to date"
				end
				next
		else
				if bVerbose
						puts "      copying #{info.source.path} => " + info.relative(webUploadDir)
				end
				system("copy \"#{info.source.path}\" \"" + info.relative(webUploadDir))
		end
}

# 8. Process the HTML files

newsItem			=	release['News-Item']
downloadsLeader		=	release['Downloads-Leader']
downloadsChanges	=	release['Downloads-Changes']
downloadsAdditions	=	release['Downloads-Additions']

# 8.a. Process the Index page

if bVerbose
		puts
		puts "  Processing the Index file"
end

indexLines				=	IO::readlines(indexPage)

insertionPointBegin		=	-1
insertionPointEnd		=	-1
insertionPointBeginVer	=	-1
insertionPointEndVer	=	-1

indexLines.each_with_index \
{ |line, lineNum_|
		lineNum = lineNum_ + 1

		#		puts "#{lineNum}: #{line.chomp}"
		
		
		# InsertionPoint-Begin
		if line =~ /<!-- #{INSERTIONPOINT_BEGIN} -->/
#				puts "#{INSERTIONPOINT_BEGIN} at line #{lineNum}"
				if insertionPointBegin >= 0
						usage("#{indexPage}(#{lineNum}): #{INSERTIONPOINT_BEGIN} already defined at line #{insertionPointBegin}")
				end
				insertionPointBegin	= lineNum_
		# InsertionPoint-End
		elsif line =~ /\<\!-- #{INSERTIONPOINT_END} --\>/
#				puts "#{INSERTIONPOINT_END} at line #{lineNum}"
				if insertionPointEnd >= 0
						usage("#{indexPage}(#{lineNum}): #{INSERTIONPOINT_END} already defined at line #{insertionPointEnd}")
				elsif insertionPointBegin < 0
						usage("#{indexPage}(#{lineNum}): #{INSERTIONPOINT_END} before #{INSERTIONPOINT_BEGIN}")
				end
				insertionPointEnd	= lineNum_
		# InsertionPoint-Begin:$(projVer)
		elsif line =~ /<!-- #{INSERTIONPOINT_BEGIN_VER} -->/
#				puts "#{INSERTIONPOINT_BEGIN_VER} at line #{lineNum}"
				if insertionPointBeginVer >= 0
						usage("#{indexPage}(#{lineNum}): #{INSERTIONPOINT_BEGIN_VER} already defined at line #{insertionPointBegin}")
				end
				insertionPointBeginVer	= lineNum_
		# InsertionPoint-End:$(projVer)
		elsif line =~ /<!-- #{INSERTIONPOINT_END_VER} -->/
#				puts "#{INSERTIONPOINT_END_VER} at line #{lineNum}"
				if insertionPointEndVer >= 0
						usage("#{indexPage}(#{lineNum}): #{INSERTIONPOINT_END_VER} already defined at line #{insertionPointEnd}")
				elsif insertionPointBeginVer < 0
						usage("#{indexPage}(#{lineNum}): #{INSERTIONPOINT_END_VER} before #{INSERTIONPOINT_BEGIN_VER}")
				end
				insertionPointEndVer	= lineNum_
		end
}

if insertionPointBegin < 0
		usage("Failed to find #{INSERTIONPOINT_BEGIN} in #{indexPage}")
end
if insertionPointEnd < 0
		usage("Failed to find #{INSERTIONPOINT_END} in #{indexPage}")
end

if ((insertionPointBeginVer + 0.5) * (insertionPointEndVer + 0.5)) < 0
		usage("Found #{INSERTIONPOINT_BEGIN_VER} and not #{INSERTIONPOINT_BEGIN_VER}, or vice versa")
end

newIndexLines		=	[]

if insertionPointEndVer < 0
		lastInsertionPointEnd	=	insertionPointEnd - 1
else
		lastInsertionPointEnd	=	insertionPointEndVer
end

newIndexLines	=	newIndexLines + indexLines[0 ... insertionPointBegin + 1]

wwwDownloadDir	=	deploy['WWWDownloadDir']
if wwwDownloadDir.length > 0 and wwwDownloadDir[wwwDownloadDir.length - 1] != '/'
		wwwDownloadDir = wwwDownloadDir + "/"
end

newIndexLines << " <!-- #{INSERTIONPOINT_BEGIN_VER} -->" << "\n"

newIndexLines << "        <p>" << "\n"
newIndexLines << "        <b>" + SrcUtil::currentDate() + "</b>: #{newsItem}" << "\n"
newIndexLines << "        <p>" << "\n"

newIndexLines << " <!-- #{INSERTIONPOINT_END_VER} -->" << "\n"

newIndexLines	=	newIndexLines + indexLines[lastInsertionPointEnd + 1 ... indexLines.length]

# 8. Update the HTML files

if bVerbose
		puts "    Writing the updated HTML file: #{indexPage}"
end

f = File::new(indexPage, "w")
newIndexLines.each { |line| f << line }
f.close()


# 8.a. Process the Downloads page

if bVerbose
		puts
		puts "  Processing the Downloads file"
end

downloadsLines			=	IO::readlines(downloadsPage)

insertionPointBegin		=	-1
insertionPointEnd		=	-1
insertionPointBeginVer	=	-1
insertionPointEndVer	=	-1

downloadsLines.each_with_index \
{ |line, lineNum_|
		lineNum = lineNum_ + 1

		# InsertionPoint-Begin
		if line =~ /<!-- #{INSERTIONPOINT_BEGIN} -->/
#				puts "#{INSERTIONPOINT_BEGIN} at line #{lineNum}"
				if insertionPointBegin >= 0
						usage("#{downloadsPage}(#{lineNum}): #{INSERTIONPOINT_BEGIN} already defined at line #{insertionPointBegin}")
				end
				insertionPointBegin	= lineNum_
		# InsertionPoint-End
		elsif line =~ /\<\!-- #{INSERTIONPOINT_END} --\>/
#				puts "#{INSERTIONPOINT_END} at line #{lineNum}"
				if insertionPointEnd >= 0
						usage("#{downloadsPage}(#{lineNum}): #{INSERTIONPOINT_END} already defined at line #{insertionPointEnd}")
				elsif insertionPointBegin < 0
						usage("#{downloadsPage}(#{lineNum}): #{INSERTIONPOINT_END} before #{INSERTIONPOINT_BEGIN}")
				end
				insertionPointEnd	= lineNum_
		# InsertionPoint-Begin:$(projVer)
		elsif line =~ /<!-- #{INSERTIONPOINT_BEGIN_VER} -->/
#				puts "#{INSERTIONPOINT_BEGIN_VER} at line #{lineNum}"
				if insertionPointBeginVer >= 0
						usage("#{downloadsPage}(#{lineNum}): #{INSERTIONPOINT_BEGIN_VER} already defined at line #{insertionPointBegin}")
				end
				insertionPointBeginVer	= lineNum_
		# InsertionPoint-End:$(projVer)
		elsif line =~ /<!-- #{INSERTIONPOINT_END_VER} -->/
#				puts "#{INSERTIONPOINT_END_VER} at line #{lineNum}"
				if insertionPointEndVer >= 0
						usage("#{downloadsPage}(#{lineNum}): #{INSERTIONPOINT_END_VER} already defined at line #{insertionPointEnd}")
				elsif insertionPointBeginVer < 0
						usage("#{downloadsPage}(#{lineNum}): #{INSERTIONPOINT_END_VER} before #{INSERTIONPOINT_BEGIN_VER}")
				end
				insertionPointEndVer	= lineNum_
		end
}

if insertionPointBegin < 0
		usage("Failed to find #{INSERTIONPOINT_BEGIN} in #{downloadsPage}")
end
if insertionPointEnd < 0
		usage("Failed to find #{INSERTIONPOINT_END} in #{downloadsPage}")
end

if ((insertionPointBeginVer + 0.5) * (insertionPointEndVer + 0.5)) < 0
		usage("Found #{INSERTIONPOINT_BEGIN_VER} and not #{INSERTIONPOINT_BEGIN_VER}, or vice versa")
end

newDownloadsLines		=	[]

if insertionPointEndVer < 0
		lastInsertionPointEnd	=	insertionPointEnd - 1
else
		lastInsertionPointEnd	=	insertionPointEndVer
end

newDownloadsLines	=	newDownloadsLines + downloadsLines[0 ... insertionPointBegin + 1]

wwwDownloadDir	=	deploy['WWWDownloadDir']
if wwwDownloadDir.length > 0 and wwwDownloadDir[wwwDownloadDir.length - 1] != '/'
		wwwDownloadDir = wwwDownloadDir + "/"
end

newDownloadsLines	<< " <!-- #{INSERTIONPOINT_BEGIN_VER} -->" << "\n"
newDownloadsLines	<< " <tr>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" colspan = \"3\">" << "\n"
newDownloadsLines	<< "   <a name = \"latest\"></a>" << "\n"
newDownloadsLines	<< "   <a name = \"#{projLabel}\"></a>" << "\n"
newDownloadsLines	<< "   <h3>version #{projVer}</h3>" << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" align = \"right\" colspan = \"1\">" << "\n"
newDownloadsLines	<< "   <b>" + SrcUtil::currentDate() + "</b>" << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< " </tr>" << "\n"

newDownloadsLines	<< " <tr>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" colspan = \"4\">" << "\n"
newDownloadsLines	<< "   #{downloadsLeader}" << "\n"
newDownloadsLines	<< "   <ul>" << "\n"
if downloadsChanges.length > 0
		newDownloadsLines	<< "    <li>" << "\n"
		newDownloadsLines	<< "     Changes / Fixes:" << "\n"
		newDownloadsLines	<< "     <ul>" << "\n"
		downloadsChanges.split(';').each \
		{ |change|
				newDownloadsLines	<< "      <li>" << "\n"
				newDownloadsLines	<< "       #{change}" << "\n"
				newDownloadsLines	<< "      </li>" << "\n"
		}
		newDownloadsLines	<< "     </ul>" << "\n"
		newDownloadsLines	<< "    </li>" << "\n"
end
if downloadsAdditions.length > 0
		newDownloadsLines	<< "    <li>" << "\n"
		newDownloadsLines	<< "     Additions:" << "\n"
		newDownloadsLines	<< "     <ul>" << "\n"
		downloadsAdditions.split(';').each \
		{ |change|
				newDownloadsLines	<< "      <li>" << "\n"
				newDownloadsLines	<< "       #{change}" << "\n"
				newDownloadsLines	<< "      </li>" << "\n"
		}
		newDownloadsLines	<< "     </ul>" << "\n"
		newDownloadsLines	<< "    </li>" << "\n"
end
newDownloadsLines	<< "   </ul>" << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< " </tr>" << "\n"


newDownloadsLines	<< " <tr>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" align = \"left\">" << "\n"
newDownloadsLines	<< "   &nbsp;" << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" align = \"left\">" << "\n"
newDownloadsLines	<< "   The recls v1.5.3 distribution." << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" align = \"left\">" << "\n"
newDownloadsLines	<< "   <a href = \"#{wwwDownloadDir}#{baseName}.zip\">#{baseName}.zip</a>" << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" align = \"left\">" << "\n"
newDownloadsLines	<< "   #{feZip.size / 1024} KB" << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< " </tr>" << "\n"

newDownloadsLines	<< " <tr>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" align = \"left\">" << "\n"
newDownloadsLines	<< "   &nbsp;" << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" align = \"left\">" << "\n"
newDownloadsLines	<< "   &nbsp;" << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" align = \"left\">" << "\n"
newDownloadsLines	<< "   <a href = \"#{wwwDownloadDir}#{baseName}.md5\">#{baseName}.md5</a>" << "\n"
newDownloadsLines	<< "    <br />" << "\n"
md5Lines	=	IO::readlines(feMD5.path)
newDownloadsLines	<< "    " + md5Lines[0].chomp
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" align = \"left\">" << "\n"
newDownloadsLines	<< "   #{feMD5.size} bytes" << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< " </tr>" << "\n"


newDownloadsLines	<< " <tr>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" align = \"left\">" << "\n"
newDownloadsLines	<< "   &nbsp;" << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" align = \"left\">" << "\n"
newDownloadsLines	<< "   STLSoft v1.8.3 beta 2 or later. Required to build recls" << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" align = \"left\">" << "\n"
newDownloadsLines	<< "   <a href = \"http://synesis.com.au/downloads/stlsoft/stlsoft-1.8.3-beta2.zip\">stlsoft-1.8.3-beta2.zip</a>" << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< "  <td valign = \"top\" align = \"left\">" << "\n"
newDownloadsLines	<< "   1075 KB" << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< " </tr>" << "\n"

 
newDownloadsLines	<< " <tr>" << "\n"
newDownloadsLines	<< "  <td colspan = \"4\">" << "\n"
newDownloadsLines	<< "   <br>" << "\n"
newDownloadsLines	<< "   <hr width = \"98%\">" << "\n"
newDownloadsLines	<< "   <br>" << "\n"
newDownloadsLines	<< "  </td>" << "\n"
newDownloadsLines	<< " </tr>" << "\n"
newDownloadsLines	<< " <!-- #{INSERTIONPOINT_END_VER} -->" << "\n"

newDownloadsLines	=	newDownloadsLines + downloadsLines[lastInsertionPointEnd + 1 ... downloadsLines.length]

if bVerbose
		puts "    Writing the updated HTML file: #{downloadsPage}"
end

f = File::new(downloadsPage, "w")
newDownloadsLines.each { |line| f << line }
f.close()

################################################################################


